import asyncio
from dataclasses import dataclass
from typing import AsyncIterable, Tuple

import faust
from http_types.types import HttpExchange
from http_types.utils import HttpExchangeBuilder

from ..meeshkan_types import HttpExchangeStream
from .abstract import AbstractSource


@dataclass(frozen=True)
class KafkaSourceConfig:
    broker: str
    topic: str


class KafkaSource(AbstractSource):
    def __init__(self, config: KafkaSourceConfig):
        self.app = faust.App(
            "meeshkan-kafka-source", broker=config.broker, stream_wait_empty=False
        )

        faust_topic = self.app.topic(config.topic, key_type=str, value_type=str)

        @self.app.agent(faust_topic)
        async def recording_agent(stream):
            """Dummy faust agent reading values from Kafka topic and passing values on.

            Yields:
                [dict] -- HttpExchange dictionary objects
            """
            async for rec in stream:
                yield rec

        self.recording_agent = recording_agent

        self.worker = None
        self.worker_task = None

    async def http_exchange_stream(self) -> AsyncIterable[HttpExchange]:
        """Generator of HttpExchange objects.
        Does not generate any values until the underlying agent is started.

        Yields:
            [HttpExchange] -- HttpExchange object
        """
        async for rec in self.recording_agent.stream():
            yield HttpExchangeBuilder.from_dict(rec)

    async def start(
        self, loop: asyncio.AbstractEventLoop
    ) -> Tuple[HttpExchangeStream, asyncio.Task]:
        self.worker = faust.Worker(self.app, loop=loop, loglevel="info")

        async def start_worker(worker: faust.Worker) -> None:
            await worker.start()

        source = self.http_exchange_stream()

        worker_coro = start_worker(self.worker)
        self.worker_task = loop.create_task(worker_coro)

        return source, self.worker_task

    def shutdown(self):
        self.worker_task.cancel()
        self.worker.stop_and_shutdown()


__all__ = ["KafkaSource", "KafkaSourceConfig"]



from meeshkan.serve.mock.storage import storage_manager


def test_clear():
    storage_manager.default["x"] = "y"
    assert storage_manager.get("default") is None

    storage_manager.add_storage("default")
    storage_manager.get("default")["x"] = "z"

    assert "y" == storage_manager.default["x"]
    assert "z" == storage_manager.get("default")["x"]

    storage_manager.clear()

    assert "x" not in storage_manager.default
    assert "z" not in storage_manager.get("default")



import pytest
from tornado.httpclient import HTTPClientError, HTTPRequest

from meeshkan.serve.mock.log import AbstractSink, Log
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app
from meeshkan.serve.mock.specs import load_specs
from meeshkan.serve.utils.routing import HeaderRouting


class MockSink(AbstractSink):
    def __init__(self):
        self.interactions = []

    def write(self, interactions):
        self.interactions = interactions


@pytest.fixture
def test_sink():
    return MockSink()


@pytest.fixture
def app(test_sink):
    return make_mocking_app(
        "tests/serve/mock/callbacks",
        load_specs("tests/serve/mock/schemas/fails"),
        HeaderRouting(),
        Log(Scope(), test_sink),
    )


@pytest.mark.gen_test
def test_mocking_server_pets(http_client, base_url, test_sink):
    req = HTTPRequest(base_url + "/", headers={"Host": "bad.api.io"})
    try:
        yield http_client.fetch(req)
        assert False  # what's a good way to do this?
    except HTTPClientError as e:
        assert str(e) == "HTTP 400: Bad Request"


@pytest.mark.gen_test
def test_mocking_server_pets_bad_path(http_client, base_url, test_sink):
    req = HTTPRequest(base_url + "/fewewsdfeaf", headers={"Host": "bad.api.io"})
    try:
        yield http_client.fetch(req)
        assert False  # what's a good way to do this?
    except HTTPClientError as e:
        assert str(e) == "HTTP 501: Not Implemented"



from hamcrest import assert_that, has_entry, instance_of, is_, matches_regexp
from http_types import RequestBuilder, Response

from meeshkan.serve.mock.matcher import truncate_path
from meeshkan.serve.mock.security import (
    match_request_to_security_scheme,
    match_to_security_schemes,
)
from meeshkan.serve.mock.specs import load_specs

spec = load_specs("tests/serve/mock/schemas/nordea")[0].api

spec_petstore = load_specs("tests/serve/mock/schemas/petstore")[0].api

redirect_uri = "https://example.com/callback"
state = "my-state"
req = RequestBuilder.from_url(
    f"https://api.nordeaopenbanking.com/personal/v4/authorize?redirect_uri={redirect_uri}&state={state}"
)


def test_truncate_path():
    truncated = truncate_path(req.pathname, spec, req)
    assert truncated == "/v4/authorize"


def test_match_to_oauth():
    match = match_request_to_security_scheme(req, spec)
    assert_that(match, instance_of(Response))


def test_match_to_security_schemes():
    match = match_to_security_schemes(req, (spec, spec_petstore))
    assert_that(match, instance_of(Response))
    assert_that(match.statusCode, is_(302))
    expected_redirect = r"^{uri}\?code=\w+".format(uri=redirect_uri)
    assert_that(match.headers, has_entry("location", matches_regexp(expected_redirect)))


def test_match_to_nonmatching_security():
    match = match_to_security_schemes(req, (spec_petstore,))
    assert_that(match, is_(None))



import json

import pytest
from tornado.httpclient import HTTPRequest

from meeshkan.serve.mock.log import AbstractSink, Log
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app
from meeshkan.serve.mock.specs import load_specs
from meeshkan.serve.utils.routing import HeaderRouting


class MockSink(AbstractSink):
    def __init__(self):
        self.interactions = []

    def write(self, interactions):
        self.interactions = interactions


@pytest.fixture
def test_sink():
    return MockSink()


@pytest.fixture
def app(test_sink):
    return make_mocking_app(
        "tests/serve/mock/callbacks",
        load_specs("tests/serve/mock/schemas/petstore"),
        HeaderRouting(),
        Log(Scope(), test_sink),
    )


URL = "petstore.swagger.io"


@pytest.mark.gen_test
def test_mocking_server_pets(http_client, base_url, test_sink):
    req = HTTPRequest(base_url + "/pets", headers={"Host": "petstore.swagger.io"})
    response = yield http_client.fetch(req)
    assert response.code == 200
    rb = json.loads(response.body)
    assert isinstance(rb, list)
    for pet in rb:
        assert isinstance(pet["name"], str)
    assert len(test_sink.interactions) == 1
    assert test_sink.interactions[0]["request"]["path"] == "/pets"


@pytest.mark.gen_test
def test_mocking_server_pet(http_client, base_url):
    req = HTTPRequest(base_url + "/pets/42", headers={"Host": "petstore.swagger.io"})
    response = yield http_client.fetch(req)
    assert response.code == 200
    rb = json.loads(response.body)
    assert isinstance(rb, dict)
    assert isinstance(rb["name"], str)
    assert isinstance(rb["id"], int)



import json
import os
from dataclasses import dataclass
from typing import Sequence, Union

import requests
import yaml
from openapi_typed_2 import OpenAPIObject, convert_to_openapi
from requests.exceptions import RequestException


@dataclass
class OpenAPISpecification:
    """An OpenAPI object with information about the source it was created from (such as the file name or URL)."""

    api: OpenAPIObject
    source: str


def load_spec(spec_source: str, is_http: bool) -> OpenAPISpecification:
    spec_text: str
    if is_http:
        try:
            response = requests.get(spec_source)
            spec_text = response.text
        except RequestException:
            raise Exception(f"Failed to load {spec_source}")
    else:
        with open(spec_source, encoding="utf8") as spec_file:
            spec_text = spec_file.read()

    return OpenAPISpecification(
        convert_to_openapi(
            (json.loads if spec_source.endswith("json") else yaml.safe_load)(spec_text)
        ),
        spec_source,
    )


def load_specs(specs: Union[str, Sequence[str]]) -> Sequence[OpenAPISpecification]:
    if isinstance(specs, str):
        specs = (specs,)

    specs_with_sources = []

    for spec in specs:
        is_http = spec.startswith("http://") or spec.startswith("https://")

        if is_http or os.path.isfile(spec):
            specs_paths = [spec]
        elif os.path.isdir(spec):
            specs_paths = []
            for dir_path, dirs, files in os.walk(spec):
                for filename in files:
                    if (
                        filename.endswith("yml")
                        or filename.endswith("yaml")
                        or filename.endswith("json")
                    ):
                        full_path = os.path.join(dir_path, filename)
                        specs_paths.append(full_path)
        else:
            raise Exception(f"OpenAPI specification {spec} not found")

        for spec_source in specs_paths:
            loaded_spec = load_spec(spec_source, is_http)
            specs_with_sources.append(loaded_spec)

    return specs_with_sources



import collections
import json
import logging
import socket
import typing
from enum import Enum
from urllib import parse

from http_types import Response
from http_types.utils import RequestBuilder
from tornado.iostream import IOStream, SSLIOStream, StreamClosedError

from meeshkan.serve.utils.routing import Routing

from ..utils.http_utils import response_from_bytes
from .proxy_callback import ProxyCallback

logger = logging.getLogger(__name__)


class ConnectionState(Enum):
    CLOSED = 0
    CONNECTING = 1
    CONNECTED = 2
    WRITING = 3


RequestInfo = collections.namedtuple(
    "RequestInfo", ["data", "scheme", "target_host", "target_port"]
)


class StreamWrapper:
    def __init__(
        self,
        stream: IOStream,
        close_callback: typing.Callable,
        queue_on_connecting: bool,
    ):
        self._queue = []
        self._stream = stream
        self._state = ConnectionState.CONNECTING

        self._stream.set_close_callback(close_callback)
        self._queue_on_connecting = queue_on_connecting

    def on_connect(self, read_callback: typing.Callable):
        self._state = ConnectionState.CONNECTED

        try:
            self._stream.read_until_close(lambda x: None, read_callback)
        except StreamClosedError:
            self._state = ConnectionState.CLOSED

    def write(self, data: bytes):
        if self._state not in {ConnectionState.CONNECTED, ConnectionState.WRITING}:
            if self._queue_on_connecting:
                self._queue.append(data)
            return

        if self._state != ConnectionState.WRITING:
            self._do_write(data)
        else:
            self._queue.append(data)

    def _do_write(self, data):
        if not data:
            self._state = ConnectionState.CLOSED
            try:
                self._stream.close()
            except StreamClosedError:
                pass
        else:
            self._client_state = ConnectionState.WRITING
            try:
                self._stream.write(data, callback=self._on_done_write())
            except StreamClosedError:
                logger.warning("Server stream closed unexpectedly")
                self._client_state = ConnectionState.CLOSED

    def _on_done_write(self):
        if self._queue:
            self._do_write(self._queue.pop(0))
            return
        self._client_state = ConnectionState.CONNECTED

    @property
    def stream(self):
        return self._stream

    @property
    def queue(self):
        return self._queue

    @property
    def state(self):
        return self._state

    def close(self):
        self._state = ConnectionState.CLOSED


class Channel:
    def __init__(
        self,
        proxy_callback: ProxyCallback,
        stream: IOStream,
        client_address: tuple,
        router: Routing,
    ):
        self._proxy_calback = proxy_callback
        self._router = router

        self._client_stream = StreamWrapper(stream, self.on_client_close, False)
        self._client_stream.on_connect(self.on_client_read)

        self._server_stream = None
        self._client_address = client_address
        self._request = None
        self._response = list()
        self._content_length = None
        self._recieved_bytes = 0

    def on_request(self, data: bytes) -> RequestInfo:
        req_lines = data.decode("utf-8").split("\r\n")
        method, fullpath, protocol = req_lines[0].split(" ")
        parsed_fullpath = parse.urlparse(fullpath)
        query = parse.parse_qs(parsed_fullpath.query)

        headers = {}
        host_line: int = 0
        body_start: int = 0
        last_line: int = 0
        for line in req_lines[1:]:
            last_line += 1
            if not line:
                break
            else:
                header, value = line.split(": ")
                if header.lower() == "host":
                    host_line = last_line
                    headers["Host"] = value
                else:
                    headers[header] = value

        body_start = last_line + 1
        body = []
        # TODO: @Nikolay, is cur_id correct?
        for body_line in range(body_start, len(req_lines)):
            if req_lines[body_line]:
                body.append(req_lines[body_line])

        body = "\r\n".join(body)

        route_info = self._router.route(parsed_fullpath.path, headers)
        fullpath = (
            "{}?{}".format(route_info.path, parsed_fullpath.query)
            if query
            else route_info.path
        )

        req_lines[0] = " ".join((method, fullpath, protocol))
        req_lines[host_line] = "Host: {}".format(route_info.host)
        headers["Host"] = route_info.host

        data = "\r\n".join(req_lines).encode("utf-8")

        # ignoring type due to this error
        """
          46:34 - error: Argument of type 'str' cannot be assigned to parameter 'method' of type 'Literal['connect', 'head', 'trace', 'options', 'delete', 'patch', 'post', 'put', 'get']'
          'str' cannot be assigned to 'Literal['connect']'
          'str' cannot be assigned to 'Literal['head']'
          'str' cannot be assigned to 'Literal['trace']'
          'str' cannot be assigned to 'Literal['options']'
          'str' cannot be assigned to 'Literal['delete']'
        """
        self._request = RequestBuilder.from_dict(
            dict(
                method=method.lower(),
                host=route_info.host,
                path=fullpath,
                pathname=route_info.path,
                protocol=route_info.scheme,
                query=query,
                body=body,
                bodyAsJson=json.loads(body) if body else {},
                headers=headers,
            )
        )

        return RequestInfo(
            data=data,
            scheme=route_info.scheme,
            target_host=route_info.hostname,
            target_port=route_info.port,
        )

    def on_response_chunk(self, data: bytes):
        if len(self._response) == 0:
            resp_lines = data.split(b"\r\n")
            for line in resp_lines[1:]:
                res = line.split(b": ")
                if len(res) == 2:
                    header, value = res[0].decode("utf-8"), res[1].decode("utf-8")
                    if header == "Content-Length":
                        self._content_length = int(value)
                        break

        self._response.append(data)
        self._recieved_bytes += len(data)

        logger.debug(
            "[%s] Received response chunk %d bytes out of %d",
            self._client_address,
            self._recieved_bytes,
            self._content_length or 0,
        )

    def flush(self, check_length=True):
        if len(self._response) > 0 and self._request is not None:
            if (
                not check_length
                or self._content_length is None
                or self._recieved_bytes >= self._content_length
            ):
                logger.debug(
                    "[%s] Flushing request, received %d bytes out of %d",
                    self._client_address,
                    self._recieved_bytes,
                    self._content_length or 0,
                )

                resp = b"".join(self._response)
                resp = response_from_bytes(resp)
                body = resp.data.decode("utf-8")

                resp = Response(
                    statusCode=resp.status,
                    body=body,
                    bodyAsJson=json.loads(body),
                    headers=dict(resp.getheaders()),
                    timestamp=None,
                )

                self._proxy_calback.on_request_complete(self._request, resp)

                self._request = None
                self._response = []
                self._content_length = None
                self._recieved_bytes = 0
        else:
            logger.debug(
                '[%s] Trying to flush incomplete data, request "%s", response chunks %d',
                self._client_address,
                str(self._request),
                len(self._response),
            )

    def on_server_read(self, data):
        self.on_response_chunk(data)
        self._client_stream.write(data)

    def on_client_read(self, data):
        self.flush()

        request_info = self.on_request(data)

        logger.debug(
            "[%s] Got request for %s://%s:%s",
            self._client_address,
            request_info.scheme,
            request_info.target_host,
            request_info.target_port,
        )

        if self._server_stream is None:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            stream = (
                SSLIOStream(s, ssl_options={})
                if request_info.scheme == "https"
                else IOStream(s)
            )
            stream.connect(
                (request_info.target_host, request_info.target_port),
                self.on_server_connect,
                server_hostname=request_info.target_host,
            )
            self._server_stream = StreamWrapper(stream, self.on_server_close, True)

        self._server_stream.write(request_info.data)

    def on_server_connect(self):
        self._server_stream.on_connect(self.on_server_read)
        if self._server_stream.queue:
            self._server_stream.write(self._server_stream.queue.pop(0))

    def remove_channel(self):
        self.flush(check_length=False)
        self._proxy_calback.on_remove_channel(self._client_address)

    def on_server_close(self):
        self._server_stream.close()
        if self._client_stream.state == ConnectionState.CLOSED:
            self.remove_channel()
        else:
            self._client_stream.write(b"")

    def on_client_close(self):
        self._client_stream.close()
        if (
            self._server_stream is None
            or self._server_stream.state == ConnectionState.CLOSED
        ):
            self.remove_channel()
        else:
            self._server_stream.write(b"")

    @property
    def server_state(self):
        return (
            ConnectionState.CLOSED
            if self._server_stream is None
            else self._server_stream.state
        )

    @property
    def client_state(self):
        return self._client_stream.state



import pytest
from hamcrest import assert_that, has_length

from meeshkan.sources.kafka import KafkaSource, KafkaSourceConfig

from ..util import read_recordings_as_dict

exchanges = read_recordings_as_dict()

kafkaSourceConfig = KafkaSourceConfig(broker="'kafka://localhost:9092", topic="example")


@pytest.fixture()
def kafka_source(event_loop):
    processor = KafkaSource(kafkaSourceConfig)
    app = processor.app
    """passing in event_loop helps avoid 'attached to a different loop' error"""
    app.finalize()
    app.conf.store = "memory://"
    app.flow_control.resume()
    return processor


@pytest.mark.asyncio()
async def test_processing(kafka_source: KafkaSource):
    async with kafka_source.recording_agent.test_context() as agent:
        kafka_source.recording_agent = agent
        for exchange in exchanges:
            await agent.put(exchange)
        res = agent.results
        assert_that(list(res.keys()), has_length(168))
        """
        # TODO How to test that kafka_source.http_exchange_stream() works without running Kafka?
        exchange_stream = kafka_source.http_exchange_stream()
        exchange_objs = [exchange async for exchange in exchange_stream]
        assert_that(exchange_objs[0], has_property("request"))
        """



import asyncio
from typing import Sequence

import click
from openapi_typed_2 import OpenAPIObject, convert_to_openapi
from yaml import safe_load

from .build.builder import BASE_SCHEMA, build_schema_async
from .build.update_mode import UpdateMode
from .config import DEFAULT_SPECS_DIR, setup
from .logger import get as getLogger
from .meeshkan_types import BuildResultStream
from .prepare import ignore_warnings
from .serve.commands import mock, record
from .sinks import AbstractSink, FileSystemSink
from .sources import AbstractSource, FileSource, KafkaSource
from .sources.kafka import KafkaSourceConfig
from .tutorial import run_tutorial

ignore_warnings()


LOGGER = getLogger(__name__)


def log(*args):
    LOGGER.debug(*args)


@click.group()
@click.version_option()
def cli():
    setup()  # Ensure setup is done before invoking the CLI.


async def write_to_sink(
    result_stream: BuildResultStream, sinks: Sequence[AbstractSink]
):
    try:
        async for result in result_stream:
            for sink in sinks:
                sink.push(result)
    finally:
        LOGGER.debug("Flushing all sinks")
        for sink in sinks:
            sink.flush()


def run_from_source(
    source: AbstractSource,
    mode: UpdateMode,
    starting_spec: OpenAPIObject,
    sinks: Sequence[AbstractSink],
):
    loop = asyncio.get_event_loop()

    if loop.is_closed():
        # Hack to allow running multiple tests that close the event loop
        loop = asyncio.new_event_loop()

    async def run(loop):
        source_stream, source_task = await source.start(loop)
        result_stream = build_schema_async(source_stream, mode, starting_spec)
        sink_task = loop.create_task(write_to_sink(result_stream, sinks))

        if source_task is not None:
            # TODO Kafka source behaves a bit bad at SIG_INT: it does not raise but returns,
            # leaving the sink task to wait for new input that never arrives. Therefore
            # one cannot `await sink_task` here, at least not without timeout
            await source_task
            try:
                await asyncio.wait_for(sink_task, timeout=3.0)
            except asyncio.TimeoutError:
                LOGGER.warn("Timeout on waiting for sink task to finish.")
        else:
            await sink_task

    try:
        loop.run_until_complete(run(loop))
    finally:
        LOGGER.info("Shutting down source")
        source.shutdown()
        loop.close()


@click.command()
@click.option(
    "-i",
    "--input-file",
    required=False,
    type=click.File("rb"),
    help="Input file. Use dash '-' to read from stdin.",
)
# TODO Isn't it a too complicated instruction? Probably, we want to change behavior to os.makedirs()
@click.option(
    "-o",
    "--out",
    required=True,
    default=DEFAULT_SPECS_DIR,
    type=click.Path(exists=False, file_okay=False, writable=True, readable=True),
    help="Output directory. If the directory does not exist, it is created if the parent directory exists.",
)
@click.option(
    "-a",
    "--initial-openapi-spec",
    required=False,
    type=click.File("rb"),
    help="Initial OpenAPI spec.",
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["GEN", "REPLAY", "MIXED"], case_sensitive=False),
    default="GEN",
    help="Spec building mode.",
)
@click.option(
    "--source",
    required=False,
    default="file",
    type=str,
    help="Source to read recordings from. For example, 'kafka'",
)
def build(input_file, out, initial_openapi_spec, mode, source):
    """
    Build an OpenAPI specification from HTTP recordings.
    """

    if input_file is not None and source != "file":
        raise Exception("Only specify input-file for --source file")

    sinks = [FileSystemSink(out)]

    if source == "kafka":
        # TODO Kafka configuration
        source = KafkaSource(
            config=KafkaSourceConfig(broker="localhost:9092", topic="http_recordings")
        )
    elif source == "file":
        if input_file is None:
            raise Exception("Option --input-file for source 'file' required.")
        source = FileSource(input_file)
    else:
        raise Exception("Unknown source {}".format(source))

    openapi_spec: OpenAPIObject = BASE_SCHEMA

    if initial_openapi_spec is not None:
        try:
            maybe_openapi = safe_load(initial_openapi_spec.read())
            # will raise if not an API spec
            openapi_spec = convert_to_openapi(maybe_openapi)
        except Exception:
            pass  # just use the initial schema
    run_from_source(source, UpdateMode[mode.upper()], openapi_spec, sinks=sinks)


@click.command()
def tutorial():
    """
    Run the Meeshkan tutorial.
    """
    run_tutorial()


cli.add_command(build)  # type: ignore
cli.add_command(record)  # type: ignore
cli.add_command(mock)  # type: ignore
cli.add_command(tutorial)  # type: ignore

if __name__ == "__main__":
    cli()



import collections
import typing
import urllib.parse

from meeshkan.serve.utils.http_utils import split_path

RoutingInfo = collections.namedtuple(
    "RoutingInfo", ["host", "hostname", "path", "scheme", "port"]
)


def get_port(scheme: str, port: typing.Optional[int]) -> int:
    if port is None:
        if scheme == "https":
            return 443
        elif scheme == "http":
            return 80
        else:
            raise AttributeError("Unsupported scheme {}".format(scheme))

    return port


class Routing:
    def route(
        self, path: str, headers: typing.Dict, inbound_scheme="http"
    ) -> RoutingInfo:
        raise NotImplementedError()


class PathRouting(Routing):
    def route(
        self, path: str, headers: typing.Dict, inbound_scheme="http"
    ) -> RoutingInfo:
        splits = split_path(path)
        url = urllib.parse.urlsplit("{}//{}".format(splits[0], splits[1]))
        path = "/" + "/".join(splits[2:])
        return RoutingInfo(
            path=path,
            host=splits[1],
            hostname=url.hostname,
            port=get_port(url.scheme, url.port),
            scheme=url.scheme,
        )


class HeaderRouting(Routing):
    def route(
        self, path: str, headers: typing.Dict, inbound_scheme="http"
    ) -> RoutingInfo:
        url = urllib.parse.urlsplit("//{}".format(headers["Host"]))
        scheme = (
            headers["X-Meeshkan-Scheme"]
            if "X-Meeshkan-Scheme" in headers
            else headers["x-meeshkan-scheme"]
            if "x-meeshkan-scheme" in headers
            else inbound_scheme
        )
        return RoutingInfo(
            path=path,
            host=headers["Host"],
            hostname=url.hostname,
            port=get_port(scheme, url.port),
            scheme=scheme,
        )


class StaticRouting(Routing):
    def __init__(self, target_address):
        self._url = urllib.parse.urlsplit(target_address)
        self._host = (
            "{}:{}".format(self._url.hostname, self._url.port)
            if self._url.port
            else self._url.hostname
        )

    def route(
        self, path: str, headers: typing.Dict, inbound_scheme="http"
    ) -> RoutingInfo:
        return RoutingInfo(
            path=path,
            host=self._host,
            hostname=self._url.hostname,
            port=get_port(self._url.scheme, self._url.port),
            scheme=self._url.scheme,
        )



########################
#### FAKER
########################
import random
from functools import reduce
from typing import Any

from faker import Faker

fkr = Faker()
__all__ = ["fake_it"]
_LO = -99999999
_HI = 99999999


# to prevent too-nested objects
def sane_depth(n):
    return max([0, 3 - n])


def fake_object(schema: Any, top_schema: Any, depth: int) -> Any:
    addls = (
        {}
        if "additionalProperties" not in schema
        else {
            k: v
            for k, v in [
                (
                    fkr.name(),
                    random.random()
                    if (isinstance(schema["additionalProperties"], bool))
                    and (schema["additionalProperties"] is True)
                    else fake_it(schema["additionalProperties"], top_schema, depth),
                )
                for x in range(random.randint(0, 4))
            ]
        }
    )
    properties = (
        [] if "properties" not in schema else [x for x in schema["properties"].keys()]
    )
    random.shuffle(properties)
    properties = (
        []
        if len(properties) == 0
        else properties[
            : min([sane_depth(depth), random.randint(0, len(properties) - 1)])
        ]
    )
    properties = list(
        set(([] if "required" not in schema else schema["required"]) + properties)
    )
    return {
        **addls,
        **{
            k: v
            for k, v in [
                (p, fake_it(schema["properties"][p], top_schema, depth))
                for p in properties
            ]
        },
    }


def fake_array(schema: Any, top_schema: Any, depth: int) -> Any:
    mn = 0 if "minItems" not in schema else schema["minItems"]
    mx = 100 if "maxItems" not in schema else schema["maxItems"]
    return (
        []
        if "items" not in schema
        else [fake_it(x, top_schema, depth) for x in schema["items"]]
        if isinstance(schema["items"], list)
        else [
            fake_it(schema["items"], top_schema, depth)
            for x in range(random.randint(mn, mx))
        ]
    )


def fake_any_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return fake_it(random.choice(schema["anyOf"]), top_schema, depth)


def fake_all_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return reduce(
        lambda a, b: {**a, **b},
        [fake_it(x, top_schema, depth) for x in schema["allOf"]],
        {},
    )


def fake_one_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return fake_it(random.choice(schema["oneOf"]), top_schema, depth)


# TODO - make this work
def fake_not(schema: Any, top_schema: Any, depth: int) -> Any:
    return {}


# TODO - make this not suck
def fake_string(schema: Any) -> str:
    return random.choice(schema["enum"]) if "enum" in schema else fkr.name()


def fake_boolean(schema: Any) -> bool:
    return (
        random.choice(schema["enum"])
        if "enum" in schema
        else True
        if random.random() > 0.5
        else False
    )


# TODO: add exclusiveMinimum and exclusiveMaximum
def fake_integer(schema: Any) -> int:
    mn = _LO if "minimum" not in schema else schema["minimum"]
    mx = _HI if "maximum" not in schema else schema["maximum"]
    return random.choice(schema["enum"]) if "enum" in schema else random.randint(mn, mx)


def fake_ref(schema: Any, top_schema, depth):
    name = schema["$ref"].split("/")[2]
    return fake_it(top_schema["definitions"][name], top_schema, depth)


def fake_null(schema: Any) -> None:
    return None


def fake_number(schema: Any) -> float:
    mn = _LO if "minimum" not in schema else schema["minimum"]
    mx = _HI if "maximum" not in schema else schema["maximum"]
    return (
        random.choice(schema["enum"])
        if "enum" in schema
        else (random.random() * (mx - mn)) + mn
    )


def fake_it(schema: Any, top_schema: Any, depth: int) -> Any:
    depth += 1
    return (
        fake_array(schema, top_schema, depth)
        if ("type" in schema) and (schema["type"] == "array")
        else fake_any_of(schema, top_schema, depth)
        if "anyOf" in schema
        else fake_all_of(schema, top_schema, depth)
        if "allOf" in schema
        else fake_one_of(schema, top_schema, depth)
        if "oneOf" in schema
        else fake_not(schema, top_schema, depth)
        if "not" in schema
        else fake_ref(schema, top_schema, depth)
        if "$ref" in schema
        else fake_object(schema, top_schema, depth)
        if ("type" not in schema) or (schema["type"] == "object")
        else fake_string(schema)
        if schema["type"] == "string"
        else fake_integer(schema)
        if schema["type"] == "integer"
        else fake_boolean(schema)
        if schema["type"] == "boolean"
        else fake_null(schema)
        if schema["type"] == "null"
        else fake_number(schema)
        if schema["type"] == "number"
        else {}
    )



from meeshkan.serve.mock.faker import fake_it
from meeshkan.serve.mock.matcher import valid_schema


def test_faker_1():
    schema = {
        "type": "array",
        "items": {
            "type": "object",
            "required": ["foo", "baz"],
            "properties": {
                "foo": {"type": "number"},
                "bar": {"type": "string"},
                "baz": {"type": "string"},
            },
        },
    }
    res = fake_it(schema, schema, 0)
    assert valid_schema(res, schema)


def test_faker_2():
    schema = {
        "$id": "https://example.com/person.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Person",
        "type": "object",
        "properties": {
            "firstName": {"type": "string", "description": "The person's first name."},
            "lastName": {"type": "string", "description": "The person's last name."},
            "age": {
                "description": "Age in years which must be equal to or greater than zero.",
                "type": "integer",
                "minimum": 0,
            },
        },
    }
    res = fake_it(schema, schema, 0)
    assert valid_schema(res, schema)


def test_faker_3():
    schema = {
        "$id": "https://example.com/geographical-location.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Longitude and Latitude Values",
        "description": "A geographical coordinate.",
        "required": ["latitude", "longitude"],
        "type": "object",
        "properties": {
            "latitude": {"type": "number", "minimum": -90, "maximum": 90},
            "longitude": {"type": "number", "minimum": -180, "maximum": 180},
        },
    }
    res = fake_it(schema, schema, 0)
    assert valid_schema(res, schema)


def test_faker_4():
    schema = {
        "$id": "https://example.com/arrays.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": "A representation of a person, company, organization, or place",
        "type": "object",
        "properties": {
            "fruits": {"type": "array", "items": {"type": "string"}},
            "vegetables": {"type": "array", "items": {"$ref": "#/definitions/veggie"}},
        },
        "definitions": {
            "veggie": {
                "type": "object",
                "required": ["veggieName", "veggieLike"],
                "properties": {
                    "veggieName": {
                        "type": "string",
                        "description": "The name of the vegetable.",
                    },
                    "veggieLike": {
                        "type": "boolean",
                        "description": "Do I like this vegetable?",
                    },
                },
            }
        },
    }
    res = fake_it(schema, schema, 0)
    assert valid_schema(res, schema)


def test_faker_5():
    schema = {"type": "array"}
    res = fake_it(schema, schema, 0)
    assert valid_schema(res, schema)



import logging
from typing import Sequence

import requests
from http_types import Request
from http_types.utils import HttpExchangeWriter
from openapi_typed_2 import convert_from_openapi, convert_to_openapi

from meeshkan.serve.mock.specs import OpenAPISpecification

logger = logging.getLogger(__name__)


class RestMiddlewareManager:
    def __init__(self):
        self._endpoints = set([])

    def get(self):
        return list(self._endpoints)

    def clear(self, url=None):
        if not url:
            self._endpoints = set([])
        else:
            self._endpoints.remove(url)

    def add(self, url):
        self._endpoints.add(url)

    def spew(
        self, request: Request, specs: Sequence[OpenAPISpecification]
    ) -> Sequence[OpenAPISpecification]:
        req = HttpExchangeWriter.to_dict(request)
        cs = {spec.source: convert_from_openapi(spec.api) for spec in specs}
        for endpoint in self._endpoints:
            res = requests.post(endpoint, json={"request": req, "schemas": cs})
            cs = res.json()
        return [
            OpenAPISpecification(convert_to_openapi(api), source)
            for source, api in cs.items()
        ]


rest_middleware_manager = RestMiddlewareManager()



"""
    isort:skip_file
"""
import asyncio
import json
import os
import shutil
import subprocess
import sys
import time
from textwrap import wrap
from urllib import request

import psutil
from pyfiglet import Figlet

from clint.textui import colored, puts
from progress.spinner import MoonSpinner

API_CALLS = """from urllib import request
from http_types import HttpExchange
from http_types.utils import RequestBuilder, ResponseBuilder, HttpExchangeWriter
from io import StringIO

def make_pokemon_request(path):
    req = request.Request('http://localhost:8000%s' % path, headers={
        'user-agent': 'python', 'Host': 'pokeapi.co', 'X-Meeshkan-Scheme': 'https'
    })
    res = request.urlopen(req)
    res.read()

PATHS = [
    '/api/v2/pokemon/1/',
    '/api/v2/pokemon/2/',
    '/api/v2/pokemon/3/',
    '/api/v2/pokemon/4/',
    '/api/v2/pokemon/5/',
    '/api/v2/pokemon/6/',
    '/api/v2/pokemon/7/',
    '/api/v2/pokemon/8/',
    '/api/v2/pokemon/9/',
    '/api/v2/pokemon/10/',
    '/api/v2/pokemon/',
    '/api/v2/type/1/',
    '/api/v2/type/2/',
    '/api/v2/type/3/',
    '/api/v2/type/4/',
    '/api/v2/type/5/',
    '/api/v2/type/6/',
    '/api/v2/type/7/',
    '/api/v2/type/8/',
    '/api/v2/type/9/',
    '/api/v2/type/10/',
    '/api/v2/type/',
    '/api/v2/ability/1/',
    '/api/v2/ability/2/',
    '/api/v2/ability/3/',
    '/api/v2/ability/4/',
    '/api/v2/ability/5/',
    '/api/v2/ability/6/',
    '/api/v2/ability/7/',
    '/api/v2/ability/8/',
    '/api/v2/ability/9/',
    '/api/v2/ability/10/',
    '/api/v2/ability/',
]

for x, path in enumerate(PATHS):
    print("  ** Calling https://pokeapi.co%s, path %d of %d" % (path, x + 1, len(PATHS)))
    make_pokemon_request(path)
"""

MERGE_SPECS = """from openapi_typed_2 import convert_to_openapi, convert_from_openapi
import json
from dataclasses import replace
import os

with open('__meeshkan__/replay/openapi.json', 'r') as replay_file:
    with open('__meeshkan__/gen/openapi.json', 'r') as gen_file:
        replay = convert_to_openapi(json.loads(replay_file.read()))
        gen = convert_to_openapi(json.loads(gen_file.read()))
        new = replace(replay, paths = { **replay.paths, **gen.paths })
        try:
            os.mkdir('__meeshkan__/both')
        except: pass # exists
        with open('__meeshkan__/both/openapi.json', 'w') as both_file:
            both_file.write(json.dumps(convert_from_openapi(new), indent=2))
"""


async def read_stream(p, server_started):
    try:
        while p.poll() is None:
            line = p.stdout.readline()
            if "Meeshkan is running" in line:
                server_started.set_result(0)
            await asyncio.sleep(0.1)
    finally:
        if not server_started.done():
            server_started.set_result(1)


async def run_bar(message, timeout, sertver_started, interval=0.1):
    bar = MoonSpinner(message)
    spent = 0
    while spent < timeout and not sertver_started.done():
        bar.next()
        await asyncio.sleep(interval)
        spent += interval
    bar.finish()


def building():
    # for now do nothing
    pass


def kill_proc_tree(p):
    if p.poll() is None:
        parent = psutil.Process(p.pid)
        children = parent.children(recursive=True)
        for child in children:
            try:
                child.kill()
            except Exception:
                pass
        psutil.wait_procs(children)
        if p.poll() is None:
            try:
                parent.kill()
            except Exception:
                pass
            parent.wait()


class CLI:
    def __init__(self, loop, use_real_input=True, throw_on_non_zero_exit=False):
        self._loop = loop
        self.use_real_input = use_real_input
        self.throw_on_non_zero_exit = throw_on_non_zero_exit

    def m_print(self, s):
        print("\n".join(wrap(s, width=60)))

    def m_input(self, s):
        text = "\n".join(wrap(s, width=60))
        if self.use_real_input:
            return input(text)
        time.sleep(2)
        return ""

    async def run(self):
        f = Figlet(font="slant")
        print(f.renderText("meeshkan"))
        puts(colored.cyan("The tutorial!!", bold=True))
        self.m_input("Press ENTER to continue...")
        ############################
        self.m_print("")
        self.m_print("##############################")
        self.m_print("")
        self.m_print(
            "Meeshkan allows you to create mocks of APIs from server traffic and OpenAPI specs.  To start, we'll record some server traffic.  But before we get started, there are a few things you should know."
        )
        self.m_print("")
        self.m_print(
            "First, Meeshkan will create a directory called __meeshkan__ in the current working directory.  Don't put anything special in there, as it may get overwritten by this tutorial!"
        )
        self.m_print("")
        self.m_print(
            "Next, this tutorial makes some network calls to the Pokemon API (pokeapi.co).  Please make sure you have a working internet connection."
        )
        self.m_print("")
        i = self.m_input(
            "With that in mind, press ENTER to continue (or the q key followed by ENTER to quit): "
        )
        if i == "q":
            self.m_print("If you change your mind, come back anytime.  Goodbye!")
            sys.exit(0)
        self.m_print("")
        self.m_print("##############################")
        self.m_print("")
        self.m_print(
            "Let's record a bit of server traffic.  We've written a file to `__meeshkan__/api_calls.py` to make our recordings.  Meeshkan expects recordings to be in the http-types format (github.com/meeshkan/http-types), so we'll use that."
        )
        self.m_print("")
        self.m_print(
            "Open up `__meeshkan__/api_calls.py`.  You'll see that we call the API 33 times using Meeshkan as a forward proxy."
        )
        self.m_print("")
        if os.path.exists("__meeshkan__"):
            shutil.rmtree("__meeshkan__")
        os.mkdir("__meeshkan__")

        with open("__meeshkan__/api_calls.py", "w") as fi:
            fi.write(API_CALLS)
        self.m_input(
            "After you've checked out `__meeshkan__/api_calls.py`, press ENTER to launch the proxy and execute the script!"
        )

        with subprocess.Popen(
            "meeshkan record -r -l __meeshkan__".split(" "),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            cwd=os.getcwd(),
            encoding="utf-8",
        ) as p:
            try:
                await self._server_starting("Starting proxy", p)

                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                res = subprocess.call("python __meeshkan__/api_calls.py", shell=True)
                if self.throw_on_non_zero_exit and (res != 0):
                    raise ValueError(
                        "Test failed at `python __meeshkan__/api_calls.py`"
                    )
                self.m_print("")
                self.m_input(
                    "Now, if you check out `__meeshkan__/pokeapi.co.jsonl`, you'll see all of the recorded server traffic. Press ENTER to continue."
                )
                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                self.m_input(
                    "The command `meeshkan build` transforms your recordings into an OpenAPI spec.  The `replay` flag tells Meeshkan to build a spec that's identical to the recorded traffic. Press ENTER to invoke `meeshkan build` in `replay` mode."
                )
                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                command = "meeshkan build -i __meeshkan__/pokeapi.co-recordings.jsonl -o __meeshkan__/replay -m replay"
                print("$ {}".format(command))
                self.m_print("")
                res = subprocess.call(
                    command,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    cwd=os.getcwd(),
                )
                if self.throw_on_non_zero_exit and (res != 0):
                    raise ValueError(f"Test failed at `{command}`")
                building()
                self.m_print("")
                self.m_print(
                    "Done.  Now, open up __meeshkan__/replay/openapi.json. Search within this document for `/api/v2/pokemon/10/:`.  This is a translation of the `GET` request you got from the Pokemon API into OpenAPI."
                )
                self.m_print("")
                self.m_input(
                    "Let's use this spec to create a server that serves back our recordings.  Press ENTER to boot up the mock server."
                )
                self.m_print("")
            finally:
                kill_proc_tree(p)

        with subprocess.Popen(
            "meeshkan mock __meeshkan__/replay -r".split(" "),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            cwd=os.getcwd(),
            encoding="utf-8",
        ) as p:
            try:
                await self._server_starting("Starting server", p)
                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                self.m_input(
                    "The server is up and running.  Press ENTER to send a `GET` request to the endpoint `/api/v2/pokemon/10/`."
                )
                req = request.Request(
                    "http://localhost:8000/api/v2/pokemon/10/",
                    headers={"Host": "pokeapi.co", "X-Meeshkan-Scheme": "https"},
                )
                res = request.urlopen(req)
                body = res.read()
                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                self.m_print("Here is the response we got back from the server.")
                self.m_print("")
                # vanilla print as thre should not be any line wraps
                # may put in function later
                print(
                    json.dumps(
                        json.loads(
                            body
                            if isinstance(body, str)
                            else body.decode("utf8")
                            if isinstance(body, bytes)
                            else ""
                        ),
                        indent=2,
                    )
                )
                self.m_print("..............................")
                self.m_print(
                    "It's the exact same response we got back from the Pokemon API.  Pretty cool, huh?"
                )
                self.m_print("")
                self.m_print(
                    "You can try the same thing.  From curl, Postman or your web browser, try calling endpoints like http://localhost:8000/api/v2/ability/ or http://localhost:8000/api/v2/type/2/.  When doing so, make sure to set the following headers:"
                )
                self.m_print("")
                print(
                    """{
    "Host": "pokeapi.co",
    "X-Meeshkan-Scheme": "https"
}"""
                )
                self.m_print("")
                self.m_input(
                    "Once you're done exploring, press ENTER to turn off the server and continue."
                )
                self.m_print("")
            finally:
                kill_proc_tree(p)

        self.m_print("")
        self.m_print("##############################")
        self.m_print("")
        self.m_print(
            "Now, let's build a new spec.  This time, instead of serving back fixed data, we will use the recordings to create _synthetic_ data.   We do this by invoking `meeshkan build --mode gen`."
        )
        self.m_print("")
        self.m_input("Press ENTER to build the new spec.")
        self.m_print("")
        self.m_print("Hang tight, we're building your spec!")
        self.m_print("")
        command = "meeshkan build -i __meeshkan__/pokeapi.co-recordings.jsonl -o __meeshkan__/gen -m gen"
        print("$ {}".format(command))
        self.m_print("")
        res = subprocess.call(
            command,
            shell=True,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            cwd=os.getcwd(),
        )
        if self.throw_on_non_zero_exit and (res != 0):
            raise ValueError(f"Test failed at `{command}`")
        self.m_print("")
        building()
        self.m_print("")
        self.m_print("Done.  In __meeshkan__/gen/, you'll see a new OpenAPI spec.")
        self.m_print("")
        self.m_input(
            "Let's use this spec to create some _synthetic_ data.  Press ENTER to reboot the mock server on port 8000."
        )
        self.m_print("")
        with subprocess.Popen(
            "meeshkan mock __meeshkan__/gen -r".split(" "),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            cwd=os.getcwd(),
            encoding="utf-8",
        ) as p:
            try:
                self.m_print("##############################")
                self.m_print("")
                await self._server_starting("Starting server", p)
                self.m_print("")
                self.m_input(
                    "The server is up and running.  Press ENTER to send a `GET` request to the endpoint `/api/v2/pokemon/10/`."
                )
                req = request.Request(
                    "http://localhost:8000/api/v2/pokemon/10/",
                    headers={"Host": "pokeapi.co", "X-Meeshkan-Scheme": "https"},
                )
                res = request.urlopen(req)
                body = res.read()
                self.m_print("")
                self.m_print("##############################")
                self.m_print("")
                self.m_print("Here is the response we got back from the server.")
                self.m_print("")
                # vanilla print as thre should not be any line wraps
                # may put in function later
                print(
                    json.dumps(
                        json.loads(
                            body
                            if isinstance(body, str)
                            else body.decode("utf8")
                            if isinstance(body, bytes)
                            else ""
                        ),
                        indent=2,
                    )
                )
                self.m_print("..............................")
                self.m_print("")
                self.m_print(
                    "The data above is synthetic, but it has the same layout as the recorded data."
                )
                self.m_print("")
                self.m_print(
                    "Why synthetic data?  Well, I'm glad you asked!  Two main reasons."
                )
                self.m_print("")
                self.m_print(
                    "1. Security breaches are most common when dealing with log files and in test environments.  So, when testing, you never want to use real data if possible."
                )
                self.m_print(
                    "2. Using synthetic data forces you write tests that focus on business logic rather than focusing on the content of fixtures, which is (in our opinion) a cleaner way to do testing."
                )
                self.m_print("")
                self.m_print(
                    "From curl, postman or your web browser, try calling http://localhost:8000/api/v2/pokemon/{id}/ , where `{id}` is _any_ positive integer. And when doing so, make sure to set following two headers:"
                )
                self.m_print("")
                print(
                    """{
    "Host": "pokeapi.co",
    "X-Meeshkan-Scheme": "https"
}"""
                )
                self.m_print("")
                self.m_input(
                    "You'll see that Meeshkan generates a synthetic response for an arbitrary Pokemon. Once you're done exploring, press ENTER to turn off the server and continue."
                )
                self.m_print("")
            finally:
                kill_proc_tree(p)

        self.m_print("")
        self.m_print("##############################")
        self.m_print("")
        with open("__meeshkan__/merge_specs.py", "w") as fi:
            fi.write(MERGE_SPECS)
        self.m_input(
            "Finally, open the file `merge_specs.py` that we created in the __meeshkan__ directory.  It's a script that merges together the two OpenAPI specs - replay and gen - created by Meeshkan.  After you've looked at it, press ENTER to execute it."
        )
        self.m_print("")
        self.m_print("$ python __meeshkan__/merge_specs.py")
        self.m_print("")
        res = subprocess.call("python __meeshkan__/merge_specs.py", shell=True)
        if self.throw_on_non_zero_exit and (res != 0):
            raise ValueError("Test failed at `python __meeshkan__/merge_specs.p`")
        self.m_print(
            "Done.  In `__meeshkan__/both/`, you'll see an OpenAPI spec that combines _both_ the fixtures from `__meeshkan__/replay/openapi.json` and the synthetic spec from `__meeshkan__/replay/openapi.json`."
        )
        self.m_print("")
        self.m_input(
            "Like the other two specs, this one can be used to create a mock server.  Try it yourself!  After this tutorial, run `meeshkan mock -i __meeshkan__/both -r`, making sure to set the same headers as before, and see how the server responds.  Press ENTER to continue."
        )
        self.m_print("")
        self.m_print("##############################")
        self.m_print("")
        self.m_print(
            "Thanks for checking out Meeshkan!  There are several other cool features, like callbacks to implement stateful logic and various connectors from libraries and platforms like Express and Kong."
        )
        self.m_print("")
        self.m_print(
            "If you have a moment, please fill out our post-tutorial survey on https://meeshkan.typeform.com/to/FpRakX.  Besides helping us improve Meeshkan, it will help us improve this and other tutorials."
        )
        self.m_print("")
        self.m_print("Take care and happy mocking!")

    async def _server_starting(self, message, p, timeout=10):
        server_started = self._loop.create_future()
        self._loop.create_task(read_stream(p, server_started))
        self._loop.create_task(run_bar(message, timeout, server_started))
        try:
            await asyncio.wait_for(server_started, timeout=timeout)
            if server_started.result() != 0:
                raise Exception(
                    "Unable to start Meeshkan. Please, check logs at ~/.meeshkan/logs for details."
                )
        except asyncio.TimeoutError:
            raise Exception(
                "Unable to start Meeshkan in 10 seconds. Please, check logs at ~/.meeshkan/logs for details."
            )


def run_tutorial():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(CLI(loop).run())



import json

import pytest
import requests_mock
from tornado.httpclient import HTTPRequest

from meeshkan.serve.mock.log import Log
from meeshkan.serve.mock.rest import rest_middleware_manager
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app
from meeshkan.serve.mock.specs import load_specs
from meeshkan.serve.utils.routing import HeaderRouting


@pytest.fixture
def app():
    return make_mocking_app(
        "tests/serve/mock/callbacks",
        load_specs("tests/serve/mock/schemas/petstore"),
        HeaderRouting(),
        Log(Scope()),
    )


@pytest.fixture(autouse=True)
def setup():
    rest_middleware_manager.add("https://api.hello.world")
    yield True
    rest_middleware_manager.clear("https://api.hello.world")


@pytest.mark.gen_test
def test_rest_middleware(http_client, base_url, app):
    req = HTTPRequest(base_url + "/pets", headers={"Host": "petstore.swagger.io"})
    with requests_mock.Mocker() as m:
        m.post(
            "https://api.hello.world",
            json={
                "petstore": {
                    "openapi": "3.0",
                    "info": {"title": "hello", "version": "world"},
                    "servers": [{"url": "http://petstore.swagger.io"}],
                    "paths": {
                        "/pets": {
                            "get": {
                                "responses": {
                                    "200": {
                                        "description": "a response",
                                        "content": {
                                            "application/json": {
                                                "schema": {
                                                    "type": "number",
                                                    "enum": [42],
                                                }
                                            }
                                        },
                                    }
                                }
                            }
                        }
                    },
                }
            },
        )
        response = yield http_client.fetch(req)
        assert 42 == json.loads(response.body)



import logging
from dataclasses import asdict
from urllib import parse

from http_types import RequestBuilder
from tornado.web import RequestHandler

from ..utils.routing import Routing
from .callbacks import CallbackManager
from .log import Log
from .response_matcher import ResponseMatcher

logger = logging.getLogger(__name__)


class MockServerView(RequestHandler):
    SUPPORTED_METHODS = ["GET", "POST", "HEAD", "DELETE", "PATCH", "PUT", "OPTIONS"]

    def initialize(
        self,
        response_matcher: ResponseMatcher,
        router: Routing,
        callback: CallbackManager,
        log: Log,
    ):
        self.response_matcher = response_matcher
        self.callback = callback
        self.router = router
        self.log = log

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, **kwargs):
        self._serve()

    def post(self):
        self._serve()

    def head(self, **kwargs):
        self._serve()

    def delete(self, **kwargs):
        self._serve()

    def patch(self, **kwargs):
        self._serve()

    def put(self, **kwargs):
        self._serve()

    def options(self, **kwargs):
        self._serve()

    def _serve(self):
        headers = {k: v for k, v in self.request.headers.get_all()}
        route_info = self.router.route(self.request.path, headers)
        headers["Host"] = route_info.host

        query = parse.parse_qs(self.request.query)
        fullpath = (
            "{}?{}".format(route_info.path, self.request.query)
            if query
            else route_info.path
        )

        # ignoring type due to this error
        """
          46:34 - error: Argument of type 'str' cannot be assigned to parameter 'method' of type 'Literal['connect', 'head', 'trace', 'options', 'delete', 'patch', 'post', 'put', 'get']'
          'str' cannot be assigned to 'Literal['connect']'
          'str' cannot be assigned to 'Literal['head']'
          'str' cannot be assigned to 'Literal['trace']'
          'str' cannot be assigned to 'Literal['options']'
          'str' cannot be assigned to 'Literal['delete']'
        """
        request = RequestBuilder.from_dict(
            {
                "method": self.request.method.lower(),
                "host": route_info.host,
                "path": fullpath,
                "pathname": route_info.path,
                "protocol": route_info.scheme,
                "query": query,
                "body": self.request.body.decode("utf-8"),
                "headers": headers,
            }
        )

        logger.debug("Processing request: %s", asdict(request))
        initial_response = self.response_matcher.get_response(request)
        response = self.callback(request, initial_response)
        logger.debug("Resolved response: %s", asdict(response))
        for header, value in response.headers.items():
            self.set_header(header, value)
        self.log.put(request, response)
        self.set_status(response.statusCode)
        self.write(response.body)
        logger.debug("Handled writing response")



from dataclasses import replace

from openapi_typed_2 import (
    OpenAPIObject,
    PathItem,
    Reference,
    convert_to_Components,
    convert_to_openapi,
    convert_to_Parameter,
    convert_to_PathItem,
    convert_to_Schema,
)

from meeshkan.serve.mock.matcher import (
    get_path_item_with_method,
    match_urls,
    matches,
    ref_name,
    use_if_header,
)


def test_match_urls():
    assert ["https://api.foo.com"] == match_urls(
        "https",
        "api.foo.com",
        convert_to_openapi(
            {
                "openapi": "",
                "paths": {},
                "info": {"title": "", "version": ""},
                "servers": [
                    {"url": "https://api.foo.commm/v2"},
                    {"url": "https://api.foo.com"},
                ],
            }
        ),
    )
    assert ["https://api.foo.com/v2/a/b/c"] == match_urls(
        "https",
        "api.foo.com",
        convert_to_openapi(
            {
                "openapi": "",
                "paths": {},
                "info": {"title": "", "version": ""},
                "servers": [
                    {"url": "https://api.foo.commm/v2"},
                    {"url": "https://api.foo.com/v2/a/b/c"},
                ],
            }
        ),
    )
    assert [] == match_urls(
        "https",
        "api.foo.com/v2",
        convert_to_openapi(
            {
                "openapi": "",
                "paths": {},
                "info": {"title": "", "version": ""},
                "servers": [
                    {"url": "https://api.foo.commm/v2"},
                    {"url": "https://api.foo.cooom/v2"},
                ],
            }
        ),
    )


def test_get_path_item_with_method():
    o: PathItem = convert_to_PathItem(
        {
            "get": {"responses": {"100": {"description": "hello"}}},
            "post": {"responses": {"101": {"description": "hello"}}},
            "delete": {"responses": {"102": {"description": "hello"}}},
            "description": "foo",
        }
    )
    assert get_path_item_with_method("get", o) == convert_to_PathItem(
        {"get": {"responses": {"100": {"description": "hello"}}}, "description": "foo"}
    )
    assert get_path_item_with_method("post", o) == convert_to_PathItem(
        {
            "post": {"responses": {"101": {"description": "hello"}}},
            "description": "foo",
        }
    )


def test_matcher():
    _bfoo = {
        "parameters": [
            {
                "in": "path",
                "name": "foo",
                "schema": {"type": "string", "pattern": "^[abc]+$"},
            },
        ],
    }
    bfoo = convert_to_PathItem(_bfoo)
    oai: OpenAPIObject = convert_to_openapi(
        {
            "openapi": "",
            "info": {"title": "", "version": ""},
            "paths": {"/b/{foo}": _bfoo},
        }
    )
    assert matches("/a/b", "/a/b", bfoo, "get", oai)
    assert not matches("/a/", "/a/b", bfoo, "get", oai)
    assert not matches("/a/b", "/a", bfoo, "get", oai)
    assert matches("/a/b/c", "/a/{fewfwef}/c", bfoo, "get", oai)
    assert matches("/b/ccaaca", "/b/{foo}", bfoo, "get", oai)
    assert not matches("/b/ccaacda", "/b/{foo}", bfoo, "get", oai)


def test_ref_name():
    assert "Foo" == ref_name(Reference(_ref="#/components/schemas/Foo", _x=None))


baseO: OpenAPIObject = convert_to_openapi(
    {
        "openapi": "hello",
        "info": {"title": "", "version": ""},
        "servers": [{"url": "https://hello.api.com"}],
        "paths": {},
    }
)


def test_use_if_header():
    assert (
        use_if_header(baseO, convert_to_Parameter({"name": "foo", "in": "query"}))
        is None
    )
    assert use_if_header(
        baseO, convert_to_Parameter({"name": "foo", "in": "header"})
    ) == ("foo", convert_to_Schema({"type": "string"}))
    assert use_if_header(
        baseO,
        convert_to_Parameter(
            {"name": "foo", "in": "header", "schema": {"type": "number"}}
        ),
    ) == ("foo", convert_to_Schema({"type": "number"}))
    assert use_if_header(
        replace(
            baseO,
            components=convert_to_Components({"schemas": {"Foo": {"type": "boolean"}}}),
        ),
        convert_to_Parameter(
            {
                "name": "foo",
                "in": "header",
                "schema": {"$ref": "#/components/schemas/Foo"},
            }
        ),
    ) == ("foo", convert_to_Schema({"type": "boolean"}))



from typing import Optional, Sequence

from http_types import Request, Response
from openapi_typed_2.openapi import OAuth2SecurityScheme, OpenAPIObject

from .matcher import truncate_path


def generate_code():
    return "xyz"


def gen_authflow_response(req: Request):
    if "redirect_uri" not in req.query:
        return Response(body="", statusCode=400, headers={})
    redirect_uri = req.query["redirect_uri"]

    redirect_uri = (
        redirect_uri if isinstance(redirect_uri, str) else "".join(redirect_uri)
    )

    code = generate_code()

    redirect_uri = f"{redirect_uri}?code={code}"

    if "state" in req.query:
        state = req.query["state"]
        redirect_uri = f"{redirect_uri}&state={state}"

    headers = dict(location=redirect_uri)
    return Response(body="", statusCode=302, headers=headers)


def matches_to_oauth2(
    req: Request, truncated_path: str, scheme: OAuth2SecurityScheme
) -> Optional[Response]:
    flows = scheme.flows

    if flows.authorizationCode is not None:
        auth_flow = flows.authorizationCode

        if truncated_path == auth_flow.authorizationUrl:
            return gen_authflow_response(req)

    # TODO Other flows
    return None


def match_request_to_security_scheme(
    req: Request, spec: OpenAPIObject
) -> Optional[Response]:
    """Match request to an OpenAPI document's security schemes if present.

    Arguments:
        req {Request} -- HttpRequest
        spec {OpenAPIObject} -- OpenAPI document

    Returns:
        Optional[Response] -- [description]
    """
    components = spec.components

    if components is None or components.securitySchemes is None:
        return None

    security_schemes = components.securitySchemes

    truncated_path = truncate_path(req.pathname, spec, req)

    for _, scheme in security_schemes.items():
        if isinstance(scheme, OAuth2SecurityScheme):
            maybe_response = matches_to_oauth2(req, truncated_path, scheme)
            if maybe_response is not None:
                return maybe_response
    return None


def match_to_security_schemes(
    req: Request, specs: Sequence[OpenAPIObject]
) -> Optional[Response]:

    matches_iterator = (
        match
        for spec in specs
        for match in (match_request_to_security_scheme(req, spec),)
        if match is not None
    )

    # Return first match when found, otherwise None
    return next(matches_iterator, None)



import logging
from typing import Optional, Sequence

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application

from ..admin.runner import start_admin
from ..mock.callbacks import CallbackManager, callback_manager
from ..mock.response_matcher import ResponseMatcher
from ..mock.specs import OpenAPISpecification
from ..mock.views import MockServerView
from ..utils.routing import PathRouting, Routing
from .log import FileSink, Log, NoSink
from .scope import Scope

logger = logging.getLogger(__name__)


def make_mocking_app_(
    callback_manager: CallbackManager,
    response_matcher: ResponseMatcher,
    router: Routing,
    log: Log,
):
    dependencies = dict(
        callback=callback_manager,
        response_matcher=response_matcher,
        router=router,
        log=log,
    )
    return Application([(r"/.*", MockServerView, dependencies)])


def make_mocking_app(
    callback_dir: Optional[str],
    specs: Sequence[OpenAPISpecification],
    routing: Routing,
    log: Log,
):
    # callback_manager = CallbackManager()
    if callback_dir is not None:
        callback_manager.load(callback_dir)

    response_matcher = ResponseMatcher(specs)

    return make_mocking_app_(callback_manager, response_matcher, routing, log)


class MockServer:
    def __init__(
        self,
        port,
        specs,
        scope: Optional[Scope] = None,
        callback_dir=None,
        admin_port=None,
        routing=PathRouting(),
        log_dir: Optional[str] = None,
    ):
        self._callback_dir = callback_dir
        self._admin_port = admin_port
        self._port = port
        self._specs = specs
        self._routing = routing
        self._scope = scope or Scope()
        self._log = Log(self._scope, NoSink() if log_dir is None else FileSink(log_dir))

    def run(self) -> None:
        if self._admin_port:
            start_admin(port=self._admin_port, scope=self._scope)
        app = make_mocking_app(
            self._callback_dir, self._specs, self._routing, self._log
        )
        http_server = HTTPServer(app)
        http_server.listen(self._port)
        self.log_startup()
        IOLoop.current().start()

    def log_startup(self) -> None:
        for spec in self._specs:
            if not spec.api.servers:
                continue
            for path, path_item in spec.api.paths.items():
                for method in [
                    "get",
                    "put",
                    "post",
                    "delete",
                    "options",
                    "head",
                    "patch",
                    "trace",
                ]:
                    if getattr(path_item, method):
                        logger.info(
                            "- "
                            + method.upper().ljust(7)
                            + f" http://localhost:{self._port}/"
                            + spec.api.servers[0].url
                            + path
                        )

        logger.info("Meeshkan is running")



import functools
import os
import sys
from logging import getLogger
from pathlib import Path

import click

from meeshkan.serve.mock.server import MockServer

from ..build.update_mode import UpdateMode
from ..config import DEFAULT_SPECS_DIR
from .mock.specs import load_specs
from .record.proxy import RecordProxyRunner
from .utils.routing import HeaderRouting, PathRouting

IS_WINDOWS = os.name == "nt"


MOCK_PID = Path.home().joinpath(".meeshkan/mock.pid")
RECORD_PID = Path.home().joinpath(".meeshkan/record.pid")
logger = getLogger(__name__)


def log_exceptions(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
            sys.exit(1)

    return wrapper


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


_common_server_options = [
    click.option("-p", "--port", default="8000", help="Server port."),
    click.option("-d", "--daemon", is_flag=True, help="Run meeshkan as a daemon."),
    click.option(
        "-r",
        "--header-routing",
        is_flag=True,
        help="Use headers to specify target hosts.",
    ),
]

_record_options = _common_server_options + [
    click.option("-l", "--log-dir", default="./logs", help="API calls logs direcotry"),
    click.option(
        "-s",
        "--specs-dir",
        default=DEFAULT_SPECS_DIR,
        help="Directory with OpenAPI specifications.",
    ),
    click.option(
        "-m",
        "--mode",
        type=click.Choice(["GEN", "REPLAY", "MIXED"], case_sensitive=False),
        default=None,
        help="Spec building mode.",
    ),
]

_mock_options = _common_server_options + [
    click.option("-a", "--admin-port", default="8888", help="Admin server port."),
    click.option(
        "-c",
        "--callback-dir",
        default=None,
        type=click.Path(exists=True, file_okay=False, resolve_path=True),
        help="Directory with callback scripts to modify behavior.",
    ),
    click.option("-k", "--kill", is_flag=True, help="Stop the mock daemon."),
    click.option(
        "-l",
        "--log-dir",
        default=None,
        type=click.Path(exists=False, file_okay=False, resolve_path=True),
        help="Directory where server logs are written.",
    ),
    click.option(
        "-s", "--status", is_flag=True, help="Show the status of the mock daemon."
    ),
]


@click.command()
@add_options(_mock_options)
@click.argument("specifications", nargs=-1)
@log_exceptions
def mock(
    callback_dir,
    log_dir,
    admin_port,
    port,
    header_routing,
    daemon,
    kill,
    status,
    specifications,
):
    """
    Run a mock server using OpenAPI specifications.
    """
    if kill:
        stop_mocking()
        sys.exit(0)
    elif status:
        status_mocking()
        sys.exit(0)

    try:
        specs = load_specs(specifications)
    except Exception as e:
        raise click.ClickException(str(e))

    server = MockServer(
        callback_dir=callback_dir,
        admin_port=admin_port,
        port=port,
        specs=specs,
        routing=HeaderRouting() if header_routing else PathRouting(),
        log_dir=log_dir,
    )

    if daemon and (not IS_WINDOWS):
        import daemonocle

        daemon = daemonocle.Daemon(
            worker=server.run, workdir=os.getcwd(), pidfile=MOCK_PID,
        )
        daemon.start()
    else:
        server.run()


def stop_mocking():
    if not IS_WINDOWS:
        import daemonocle

        daemon = daemonocle.Daemon(pidfile=MOCK_PID,)
        daemon.stop()


def status_mocking():
    if not IS_WINDOWS:
        import daemonocle

        daemon = daemonocle.Daemon(pidfile=MOCK_PID,)
        daemon.status()


@click.group(invoke_without_command=True)
@add_options(_record_options)
@click.pass_context
@log_exceptions
def record(ctx, port, log_dir, header_routing, specs_dir, mode, daemon):
    """
    Record HTTP traffic from a reverse proxy.
    """
    if ctx.invoked_subcommand is None:
        ctx.forward(start_record)


@record.command(name="start")  # type: ignore
@add_options(_record_options)
@log_exceptions
def start_record(port, log_dir, header_routing, specs_dir, mode, daemon):
    print(os.getpid())
    proxy_runner = RecordProxyRunner(
        port=port,
        log_dir=log_dir,
        routing=HeaderRouting() if header_routing else PathRouting(),
        specs_dir=specs_dir,
        mode=UpdateMode[mode.upper()] if mode else None,
    )
    if daemon and (not IS_WINDOWS):
        import daemonocle

        daemon = daemonocle.Daemon(
            worker=proxy_runner.run, workdir=os.getcwd(), pidfile=RECORD_PID,
        )
        daemon.start()
    else:
        proxy_runner.run()


@record.command(name="stop")  # type: ignore
def stop_recording():
    if not IS_WINDOWS:
        import daemonocle

        daemon = daemonocle.Daemon(pidfile=RECORD_PID,)
        daemon.stop()


@record.command(name="status")  # type: ignore
def status_recording():
    if not IS_WINDOWS:
        import daemonocle

        daemon = daemonocle.Daemon(pidfile=RECORD_PID,)
        daemon.status()



import json
import os
import time
from dataclasses import dataclass
from typing import Optional, Sequence

from http_types import HttpExchange, HttpExchangeWriter, Request, Response

from .scope import Scope


@dataclass
class MeeshkanMeta:
    timestamp: int
    scope: Optional[str]


@dataclass
class LoggedHttpExchange:
    request: Request
    response: Response
    meta: MeeshkanMeta


class AbstractSink:
    def write(self, interactions):
        pass


class NoSink(AbstractSink):
    def write(self, interactions):
        pass


class FileSink(AbstractSink):
    def __init__(self, log_dir: str):
        self._dir = log_dir
        if not os.path.exists(self._dir):
            os.mkdir(self._dir)
        self._log_file_name = "%d.log.json" % int(time.time() * 1000)

    def write(self, interactions):
        with open(os.path.join(self._dir, self._log_file_name), "w") as logfile:
            logfile.write(json.dumps(interactions, indent=2))


class Log:
    _interactions: Sequence[LoggedHttpExchange]

    def __init__(self, scope: Scope, sink: Optional[AbstractSink] = None):
        self._scope = scope
        self._sink = sink
        self._interactions = []
        self._interactions_as_json = []
        self._log_file_name = "%d.log.json" % int(time.time() * 1000)

    def put(self, request: Request, response: Response):
        self._interactions = [
            *self._interactions,
            LoggedHttpExchange(
                request=request,
                response=response,
                meta=MeeshkanMeta(
                    timestamp=int(time.time() * 1000), scope=self._scope.get()
                ),
            ),
        ]
        exchange = HttpExchangeWriter.to_dict(
            HttpExchange(request=request, response=response,)
        )
        interaction = {
            **exchange,
            "meta": {
                "timestamp": self._interactions[-1].meta.timestamp,
                **(
                    {"scope": self._interactions[-1].meta.scope}
                    if self._interactions[-1].meta.scope is not None
                    else {}
                ),
            },
        }

        self._interactions_as_json = [*self._interactions_as_json, interaction]
        if self._sink is not None:
            self._sink.write(self._interactions_as_json)



from typing import Sequence

import requests_mock

from meeshkan.serve.mock.specs import OpenAPISpecification, load_specs

spec_directory = "tests/serve/mock/schemas/petstore"
spec_file_path = f"{spec_directory}/index.yaml"


def assert_on_petstore_spec(
    specs: Sequence[OpenAPISpecification], expected_source: str
):
    assert len(specs) == 1
    spec = specs[0]
    assert expected_source == spec.source.replace("\\", "/")  # for windows
    assert "Swagger Petstore" == spec.api.info.title


def test_load_specs_from_dir():
    loaded_specs = load_specs((spec_directory,))
    assert_on_petstore_spec(loaded_specs, spec_file_path)


def test_load_specs_from_file():
    loaded_specs = load_specs((spec_file_path,))
    assert_on_petstore_spec(loaded_specs, spec_file_path)


def test_load_specs_from_http():
    with open(spec_file_path, encoding="utf8") as f:
        spec_text = f.read()

    for protocol in ["http", "https"]:
        spec_url = f"{protocol}://api.example.com/index.yaml"
        with requests_mock.Mocker() as m:
            m.get(spec_url, text=spec_text)
            loaded_specs = load_specs(spec_url)
            assert_on_petstore_spec(loaded_specs, spec_url)



from http_types.utils import RequestBuilder, ResponseBuilder

from meeshkan.serve.mock.callbacks import callback_manager
from meeshkan.serve.mock.storage import storage_manager


def test_no_callback():
    callback_manager.load("tests/serve/mock/callbacks")

    request = RequestBuilder.from_dict(
        dict(
            method="get",
            host="api.com",
            pathname="/nothing",
            query={},
            body="",
            protocol="http",
            bodyAsJson={},
            headers={},
        )
    )

    response = ResponseBuilder.from_dict(
        dict(statusCode=200, body="", bodyAsJson={}, headers={})
    )

    new_response = callback_manager(request, response)

    assert response == new_response


def test_json():
    callback_manager.load("tests/serve/mock/callbacks")
    storage_manager.clear()

    request = RequestBuilder.from_dict(
        dict(
            method="post",
            host="api.com",
            pathname="/counter",
            query={},
            body="",
            protocol="http",
            bodyAsJson={},
            headers={},
        )
    )

    response = ResponseBuilder.from_dict(
        dict(statusCode=200, body="", bodyAsJson={"field": "value"}, headers={})
    )

    new_response = callback_manager(request, response)

    assert 1 == new_response.bodyAsJson["count"]
    assert "value" == new_response.bodyAsJson["field"]
    assert "count" in new_response.body

    request_set = RequestBuilder.from_dict(
        dict(
            method="post",
            host="api.com",
            pathname="/counter",
            query={},
            body="",
            protocol="http",
            bodyAsJson={"set": 10},
            headers={},
        )
    )

    new_response = callback_manager(request_set, response)

    assert 10 == new_response.bodyAsJson["count"]
    assert "value" == new_response.bodyAsJson["field"]
    assert "count" in new_response.body

    new_response = callback_manager(request, response)
    assert 11 == new_response.bodyAsJson["count"]
    assert "value" == new_response.bodyAsJson["field"]
    assert "count" in new_response.body


def test_text():
    callback_manager.load("tests/serve/mock/callbacks")
    storage_manager.clear()

    request = RequestBuilder.from_dict(
        dict(
            method="get",
            host="api.com",
            pathname="/text_counter",
            query={"set": 10},
            body="",
            protocol="http",
            bodyAsJson={},
            headers={},
        )
    )

    response = ResponseBuilder.from_dict(
        dict(statusCode=200, body="Called", bodyAsJson={}, headers={})
    )

    new_response = callback_manager(request, response)

    assert 10 == new_response.headers["x-meeshkan-counter"]
    assert "Called 10 times" == new_response.body


def test_json_full():
    callback_manager.load("tests/serve/mock/callbacks")
    storage_manager.clear()

    request = RequestBuilder.from_dict(
        dict(
            method="get",
            host="another.api.com",
            pathname="/echo",
            query={"message": "Hello"},
            body="",
            protocol="http",
            bodyAsJson={},
            headers={},
        )
    )

    response = ResponseBuilder.from_dict(
        dict(statusCode=200, body="", bodyAsJson={"field": "value"}, headers={})
    )

    new_response = callback_manager(request, response)

    assert "Hello" == new_response.bodyAsJson["message"]
    assert "value" == new_response.headers["X-Echo-Header"]


def test_text_full():
    callback_manager.load("tests/serve/mock/callbacks")
    storage_manager.clear()

    request = RequestBuilder.from_dict(
        dict(
            method="post",
            host="another.api.com",
            pathname="/echo",
            query={},
            protocol="http",
            body="Hello",
            bodyAsJson={},
            headers={},
        )
    )

    response = ResponseBuilder.from_dict(
        dict(statusCode=200, body="", bodyAsJson={"field": "value"}, headers={})
    )

    new_response = callback_manager(request, response)

    assert "Hello" == new_response.body
    assert "value" == new_response.headers["X-Echo-Header"]



import glob
import importlib.util
import inspect
import json
import logging
import os
from dataclasses import asdict

from http_types.utils import ResponseBuilder

from .storage import storage_manager

logger = logging.getLogger(__name__)


class CallbackManager:
    ARGS_MAP_COMMON = {
        "query": lambda request, response, storage: request["query"],
        "request_headers": lambda request, response, storage: request["headers"],
        "response_headers": lambda request, response, storage: response["headers"],
        "request": lambda request, response, storage: request,
        "response": lambda request, response, storage: response,
        "storage": lambda request, response, storage: storage,
    }

    ARGS_MAP_JSON = {
        **ARGS_MAP_COMMON,
        "request_body": lambda request, response, storage: request["bodyAsJson"],
        "response_body": lambda request, response, storage: response["bodyAsJson"],
    }

    ARGS_MAP_TEXT = {
        **ARGS_MAP_COMMON,
        "request_body": lambda request, response, storage: request["body"],
        "response_body": lambda request, response, storage: response["body"],
    }

    def __init__(self):
        self._callbacks = dict()

    def load(self, path):
        if not os.path.isdir(path):
            raise FileNotFoundError(
                f"Callback configuration directory doesn't exist: {path}"
            )

        for f in glob.glob(os.path.join(path, "*.py")):
            module_name = "callbacks.{}".format(
                os.path.splitext(os.path.basename(f))[0]
            )
            logging.debug("Loading callbacks from %s to %s module", f, module_name)
            # from https://stackoverflow.com/questions/19009932/import-arbitrary-python-source-file-python-3-3
            spec = importlib.util.spec_from_file_location(module_name, f)
            callbacks = importlib.util.module_from_spec(spec)
            # even though this is an accepted response, the line below
            # triggers a typing error. ignore types for now, but should
            # investigate why the type does not work
            spec.loader.exec_module(callbacks)  # type: ignore

        logging.info("Loaded %d callbacks", len(self._callbacks))

    def callback(self, host, method, path):
        return self._callbacks.get((host, method, path))

    def add_callback(self, host, method, path, format, response_type, callback):
        args = inspect.getfullargspec(callback).args
        args_map = self.ARGS_MAP_JSON if format == "json" else self.ARGS_MAP_TEXT
        self._callbacks[
            (host, method, path)
        ] = lambda request, response, storage: self._map_results(
            response,
            callback(
                **{arg: args_map[arg](request, response, storage) for arg in args}
            ),
            format,
            response_type,
        )

    def _map_results(self, response, result, format, response_type):
        if response_type == "body":
            if format == "json":
                response["bodyAsJson"] = result
                response["body"] = json.dumps(result)
            else:
                response["body"] = result
        else:
            if format == "json":
                response["body"] = json.dumps(result)

        return response

    def __call__(self, request, response):
        callback = self._callbacks.get(
            (request.host, request.method.value, request.pathname)
        )
        if callback is not None:
            out = callback(asdict(request), asdict(response), storage_manager.default)
            return ResponseBuilder.from_dict(out)
        else:
            return response


callback_manager = CallbackManager()


def callback(host, method, path, format="json", response="body"):
    def decorator(function):
        callback_manager.add_callback(host, method, path, format, response, function)
        return function

    return decorator



import json
import logging

from tornado.web import RequestHandler

from ..mock.rest import RestMiddlewareManager
from ..mock.scope import Scope
from ..mock.storage import StorageManager

logger = logging.getLogger(__name__)


class StorageView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE"]

    def initialize(self, storage_manager: StorageManager):
        self.storage_manager = storage_manager

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self):
        self.storage_manager.clear()


class ScopeView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE", "POST", "GET"]

    def initialize(self, scope: Scope):
        self.scope = scope

    def delete(self):
        self.scope.clear()

    def post(self):
        self.scope.set(self.get_body_argument("name"))

    def get(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')
        return self.write(
            dict() if self.scope.get() is None else dict(name=self.scope.get())
        )


class RestMiddlewaresView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE", "GET"]

    def initialize(self, rest_middleware_manager: RestMiddlewareManager):
        self.rest_middleware_manager = rest_middleware_manager

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self):
        self.rest_middleware_manager.clear()

    def get(self):
        self.write(json.dumps(self.rest_middleware_manager.get()))


class RestMiddlewareView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE", "POST"]

    def initialize(self, rest_middleware_manager: RestMiddlewareManager):
        self.rest_middleware_manager = rest_middleware_manager

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self, url):
        self.rest_middleware_manager.clear(url)

    def post(self, url):
        self.rest_middleware_manager.add(url)



import json

import pytest

from meeshkan.serve.mock.log import Log
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app
from meeshkan.serve.mock.specs import load_specs
from meeshkan.serve.utils.routing import PathRouting


@pytest.fixture
def app():
    return make_mocking_app(
        "tests/serve/mock/callbacks",
        load_specs("tests/serve/mock/schemas/stripe"),
        PathRouting(),
        Log(Scope()),
    )


@pytest.mark.gen_test(timeout=10)
# @pytest.mark.skip("takes too long due to a timeout, need to investigate")
def test_mocking_server_customers(http_client, base_url):
    response = yield http_client.fetch(
        base_url + "/https://api.stripe.com/v1/customers"
    )
    assert response.code == 200
    rb = json.loads(response.body)
    assert isinstance(rb, dict)



import asyncio
from typing import AsyncIterable, Awaitable, Callable, Tuple

from http_types import HttpExchange

from .build.result import BuildResult

HttpExchangeStream = AsyncIterable[HttpExchange]
BuildResultStream = AsyncIterable[BuildResult]

Source = Callable[[asyncio.AbstractEventLoop], Tuple[HttpExchangeStream, asyncio.Task]]
Sink = Callable[[BuildResultStream], Awaitable[None]]



import asyncio
from typing import AsyncIterable, TextIO, Tuple

from http_types.types import HttpExchange
from http_types.utils import HttpExchangeReader

from .abstract import AbstractSource


class FileSource(AbstractSource):
    def __init__(self, input_file: TextIO):
        self.input_file = input_file

    async def start(
        self, loop: asyncio.AbstractEventLoop
    ) -> Tuple[AsyncIterable[HttpExchange], None]:
        # print(self.input_file)

        async def read():
            for line in self.input_file:
                if line.rstrip() not in [b"", ""]:  # prevent reading empty space
                    yield HttpExchangeReader.from_json(line)

        return read(), None

    def shutdown(self):
        pass



import json
import logging

import pytest
from hamcrest import assert_that, has_key, is_
from tornado.httpclient import HTTPClientError, HTTPRequest

from meeshkan.serve.mock.log import Log
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app
from meeshkan.serve.mock.specs import load_specs
from meeshkan.serve.utils.routing import PathRouting

logger = logging.getLogger(__name__)
logging.basicConfig(level="DEBUG")

routing = PathRouting()
log = Log(Scope())
specs = load_specs("tests/serve/mock/schemas/nordea")


@pytest.fixture
def app():
    return make_mocking_app(None, specs, PathRouting(), log,)


def make_sandbox_url(base_url, path: str):
    return f"{base_url}/https://api.nordeaopenbanking.com{path}"


class TestNordea:
    @pytest.mark.gen_test
    async def test_nordea_authorize_returns_302(self, http_client, base_url):
        redirect_uri = "http://httpbin.org/get"
        query = f"state=fake&client_id=CLIENT_ID&scope=ACCOUNTS_BASIC,ACCOUNTS_BALANCES,ACCOUNTS_DETAILS,ACCOUNTS_TRANSACTIONS,PAYMENTS_MULTIPLE,CARDS_INFORMATION,CARDS_TRANSACTIONS&duration=129600&redirect_uri={redirect_uri}&country=FI"
        url = make_sandbox_url(base_url, path=f"/personal/v4/authorize?{query}")
        req = HTTPRequest(url, follow_redirects=False)
        with pytest.raises(HTTPClientError) as e:  # type: ignore
            await http_client.fetch(req)
        assert_that(e.value.code, is_(302))

    @pytest.mark.skip("Returning 401 for invalid auth not implemented")
    @pytest.mark.gen_test
    async def test_nordea_accounts_returns_401(self, http_client, base_url):
        url = make_sandbox_url(base_url, path="/personal/v4/accounts")
        req = HTTPRequest(url)
        try:
            await http_client.fetch(req)
            pytest.fail("Expected failure without auth")
        except HTTPClientError as e:
            assert e.code == 401, "Expected 401 without auth"

    @pytest.mark.skip("Returns randomly different status codes for a valid request")
    @pytest.mark.gen_test
    async def test_nordea_accounts(self, http_client, base_url):
        headers = {
            "Signature": "fake-signature",
            "X-IBM-Client-ID": "fake-client-id",
            "X-IBM-Client-Secret": "fake-client-secret",
            "X-Nordea-Originating-Date": "blah",
            "X-Nordea-Originating-Host": "blah2",
        }
        url = make_sandbox_url(base_url, path="/personal/v4/accounts")
        req = HTTPRequest(url, headers=headers)
        response = await http_client.fetch(req)
        assert response.code == 200, "Expected 200 for a valid request"
        rb = json.loads(response.body)
        assert isinstance(rb, dict)
        assert_that(rb, has_key("group_header"))
        assert_that(rb, has_key("response"))



class Scope:
    def __init__(self):
        self._name = None

    def set(self, name):
        self._name = name

    def get(self):
        return self._name

    def clear(self):
        self._name = None



import warnings


def ignore_warnings():
    """Several of our deps have deprecation warnings
    # we ignore for now, but should look into."""
    ### This comes from faust """
    warnings.filterwarnings(
        action="ignore", category=DeprecationWarning, module=r"venusian"
    )
    ### This comes from lenses
    warnings.filterwarnings(
        action="ignore", category=DeprecationWarning, module=r"singledispatch_helpers"
    )



import json
import logging
import random
from typing import Mapping, Sequence, Union, cast

from faker import Faker
from http_types import Request, Response
from openapi_typed_2 import Reference, convert_from_openapi

from .faker import fake_it
from .matcher import (
    change_ref,
    change_refs,
    get_response_from_ref,
    match_request_to_openapi,
    ref_name,
)
from .rest import rest_middleware_manager
from .security import match_to_security_schemes
from .specs import OpenAPISpecification

fkr = Faker()


logger = logging.getLogger(__name__)


class ResponseMatcher:
    _specs: Sequence[OpenAPISpecification]

    def __init__(self, specs: Sequence[OpenAPISpecification]):
        self._specs = specs

    def match_error(self, msg: str, req: Request) -> Response:
        return self.default_response(
            "%s. Here is the full request: host=%s, path=%s, method=%s."
            % (msg, req.host, req.path, req.method.value)
        )

    def default_response(self, msg):
        json_resp = {"message": msg}
        return Response(
            statusCode=501,
            body=json.dumps(json_resp),
            bodyAsJson=json_resp,
            headers={},
            timestamp=None,
        )

    def get_response(self, request: Request) -> Response:
        # TODO: tight coupling here
        # try to decouple...
        specs = rest_middleware_manager.spew(request, self._specs)

        logger.debug("Matching to security schemes of %d specs", len(specs))
        maybe_security_response = match_to_security_schemes(
            request, [spec.api for spec in specs]
        )

        if maybe_security_response is not None:
            logger.debug("Matched to security scheme, returning response.")
            return maybe_security_response

        logger.debug("Matching to paths")

        matches = match_request_to_openapi(request, specs)

        logger.debug("Found %d matching specs", len(matches))
        if len(matches) == 0:
            return self.match_error(
                "Could not find an OpenAPI schema for the host %s." % request.host,
                request,
            )

        match = random.choice(matches)

        logger.debug("Proceeding with document title %s", match.api.info.title)

        path_error = "Could not find a path %s on hostname %s." % (
            request.path,
            request.host,
        )

        if match.api.paths is None:
            return self.match_error(path_error, request)
        if len(match.api.paths.items()) == 0:
            return self.match_error(path_error, request)

        path_candidates = [x for x in match.api.paths.values()]

        path_candidate = random.choice(path_candidates)

        method = {
            "get": path_candidate.get,
            "post": path_candidate.post,
            "put": path_candidate.put,
            "delete": path_candidate.delete,
            "options": path_candidate.options,
            "head": path_candidate.head,
            "patch": path_candidate.patch,
            "trace": path_candidate.trace,
        }.get(str(request.method.value), None)

        method_error = "Could not find a method %s for path %s on hostname %s." % (
            request.method.value,
            request.path,
            request.host,
        )

        if method is None:
            return self.match_error(method_error, request)
        responses_error = "While a stub for a specification exists for this endpoint, it contains no responses. That usually means the schema is corrupt or it has been constrained too much (ie asking for a 201 response when it only has 200 and 400)."
        if method.responses is None:
            return self.match_error(responses_error, request)
        potential_responses = [r for r in method.responses.items()]

        if len(potential_responses) == 0:
            return self.match_error(responses_error, request)

        response = random.choice(potential_responses)
        response = potential_responses[0]
        response_1 = response[1]
        response_1 = (
            get_response_from_ref(match.api, ref_name(response_1))
            if isinstance(response_1, Reference)
            else response_1
        )
        if response_1 is None:
            return self.match_error(responses_error, request)
        headers: Mapping[str, Union[str, Sequence[str]]] = {}
        if response_1.headers is not None:
            # TODO: can't handle references yet, need to fix
            headers = (
                {}
            )  # { k: (faker(v['schema'], v['schema'], 0) if 'schema' in v else '***') for k,v in headers.items() }
        statusCode = int(response[0] if response[0] != "default" else 400)
        if (response_1.content is None) or len(response_1.content.items()) == 0:
            return Response(
                statusCode=statusCode,
                body="",
                headers=headers,
                bodyAsJson=None,
                timestamp=None,
            )
        mime_types = response_1.content.keys()
        if "application/json" in mime_types:
            content = response_1.content["application/json"]
            if content.schema is None:
                return self.match_error("Could not find schema", request)
            schema = content.schema
            ct: Mapping[str, Union[str, Sequence[str]]] = {
                "Content-Type": "application/json"
            }
            new_headers: Mapping[str, Union[str, Sequence[str]]] = {**headers, **ct}
            if schema is None:
                return Response(
                    statusCode=statusCode,
                    body="",
                    bodyAsJson="",
                    headers=new_headers,
                    timestamp=None,
                )
            to_fake = {
                **convert_from_openapi(
                    change_ref(schema)
                    if isinstance(schema, Reference)
                    else change_refs(schema)
                ),
                "definitions": {
                    k: convert_from_openapi(
                        change_ref(v) if isinstance(v, Reference) else change_refs(v)
                    )
                    for k, v in (
                        match.api.components.schemas.items()
                        if (match.api.components is not None)
                        and (match.api.components.schemas is not None)
                        else []
                    )
                },
            }
            bodyAsJson = fake_it(to_fake, to_fake, 0)
            return Response(
                statusCode=statusCode,
                body=json.dumps(bodyAsJson),
                bodyAsJson=bodyAsJson,
                headers=new_headers,
                timestamp=None,
            )
        if "text/plain" in mime_types:
            return Response(
                statusCode=statusCode,
                body=fkr.sentence(),
                # TODO: can this be accomplished without a cast?
                headers=cast(
                    Mapping[str, Union[str, Sequence[str]]],
                    {**headers, "Content-Type": "text/plain"},
                ),
                bodyAsJson=None,
                timestamp=None,
            )
        return self.match_error(
            "Could not produce content for these mime types %s" % str(mime_types),
            request,
        )



import os
import sys
from shutil import rmtree

from setuptools import Command, errors, find_packages, setup

# Package meta-data.
NAME = "meeshkan"
DESCRIPTION = "Reverse engineer services with style"
URL = "http://github.com/meeshkan/meeshkan"
EMAIL = "dev@meeshkan.com"
AUTHOR = "Meeshkan Dev Team"
REQUIRES_PYTHON = ">=3.6.0"

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

REQUIRED = [
    "click==7.0",
    "lenses",
    "pyyaml",
    "jsonschema",
    "dataclasses",  # for 3.6, as it ships with 3.7
    "faker",
    "requests",
    "typing-extensions",
    "openapi-typed_2>=0.0.2",
    "typeguard>=2.7.0",
    "genson",
    "http-types>=0.0.14,<0.1.0",
    # kafka
    "faust",
    # mock
    "tornado==5.1.1",
    "urllib3==1.25.6",
    "meeshkan-daemonocle",
    # tutorial
    "clint",
    "pyfiglet",
    "progress",
    "psutil>=5.6.3",
]

BUNDLES = {}

# Requirements of all bundles
BUNDLE_REQUIREMENTS = [dep for _, bundle_dep in BUNDLES.items() for dep in bundle_dep]

DEV = BUNDLE_REQUIREMENTS + [
    "black==19.10b0",
    "flake8",
    "isort",
    "mypy",
    "pyhamcrest",
    "pylint",
    "pytest",
    "pytest-asyncio",
    "pytest-testmon",
    "pytest-tornado",
    "pytest-watch",
    "requests-mock",
    "setuptools",
    "twine",
    "wheel",
]

VERSION = "0.2.23"

ENTRY_POINTS = ["meeshkan = meeshkan.__main__:cli"]

EXTRAS = dict(**BUNDLES, dev=DEV)


def run_sys_command(cmd, error_msg):
    """Run command in os.system(), raise if exit_code non-zero.

    Arguments:
        cmd {[type]} -- Command to run. For example: "pytest"
        error_msg {[type]} -- Error message to raise at failure

    Raises:
        errors.DistutilsError: Exited with non-zero exit code.
    """
    exit_code = os.system(cmd)
    if exit_code != 0:
        raise errors.DistutilsError(error_msg)


class SetupCommand(Command):
    """Base class for setup.py commands with no arguments"""

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def rmdir_if_exists(self, directory):
        self.status("Deleting {}".format(directory))
        rmtree(directory, ignore_errors=True)


BUILD_COMMAND = "{executable} setup.py sdist bdist_wheel --universal".format(
    executable=sys.executable
)

TYPE_CHECK_COMMAND = "pyright --lib"

TEST_COMMAND = "pytest"

LINT_COMMAND = "flake8 --exclude .git,.venv,__pycache__,build,dist"

BLACK_FORMAT_COMMAND = "black ."
ISORT_FORMAT_COMMAND = "isort -y"
BLACK_CHECK_COMMAND = "black --check ."
ISORT_CHECK_COMMAND = "isort --check-only"


def build():
    run_sys_command(BUILD_COMMAND, "Build failed")


def type_check():
    run_sys_command(TYPE_CHECK_COMMAND, "Type-checking failed")


def run_tests():
    run_sys_command(TEST_COMMAND, "Tests failed")


def check_style():
    run_sys_command(LINT_COMMAND, "Checking style failed")


def enforce_formatting():
    run_sys_command(ISORT_FORMAT_COMMAND, "Formatting with isort failed")
    run_sys_command(BLACK_FORMAT_COMMAND, "Formatting with black failed")


def check_formatting():
    run_sys_command(ISORT_CHECK_COMMAND, "Checking with isort failed")
    run_sys_command(BLACK_CHECK_COMMAND, "Checking with black failed")


class BuildDistCommand(SetupCommand):
    """Support setup.py upload."""

    description = "Build the package."

    def run(self):
        self.status("Removing previous builds...")
        self.rmdir_if_exists(os.path.join(here, "dist"))
        self.status("Building Source and Wheel (universal) distribution...")
        build()


class FormatCommand(SetupCommand):
    """Enforce formatting."""

    description = "Enforce formatting."

    def run(self):
        enforce_formatting()


class TypeCheckCommand(SetupCommand):
    """Run type-checking."""

    description = "Run type-checking."

    def run(self):
        type_check()


class TestCommand(SetupCommand):
    """Support setup.py test."""

    description = "Run tests, formatting, type-checks, and linting"

    def run(self):
        self.status("Checking formatting...")
        check_formatting()

        self.status("Checking style...")
        check_style()

        self.status("Checking types...")
        type_check()

        self.status("Running pytest...")
        run_tests()


class UploadCommand(SetupCommand):
    """Support setup.py upload."""

    description = "Build and publish the package."

    def run(self):

        self.status("Removing previous builds...")
        self.rmdir_if_exists(os.path.join(here, "dist"))

        self.status("Building Source and Wheel (universal) distribution...")
        build()

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        self.status("Pushing git tags...")
        os.system("git tag v{about}".format(about=VERSION))
        os.system("git push --tags")

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    entry_points={"console_scripts": ENTRY_POINTS},
    cmdclass={
        "dist": BuildDistCommand,
        "format": FormatCommand,
        "upload": UploadCommand,
        "test": TestCommand,
        "typecheck": TypeCheckCommand,
    },
)



import json
import re
from dataclasses import replace
from functools import reduce
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from urllib.parse import urlparse

import jsonschema
import lenses
from http_types import Request
from lenses import lens
from openapi_typed_2 import (
    Components,
    Header,
    MediaType,
    OpenAPIObject,
    Operation,
    Parameter,
    PathItem,
    Reference,
    RequestBody,
    Response,
    Responses,
    Schema,
    convert_from_openapi,
)

from meeshkan.build.operation import operation_from_string
from meeshkan.serve.mock.specs import OpenAPISpecification

C = TypeVar("C")
D = TypeVar("D")
S = TypeVar("S")
T = TypeVar("T")
W = TypeVar("W")
X = TypeVar("X")
Y = TypeVar("Y")
Z = TypeVar("Z")

__all__ = ["match_request_to_openapi", "change_ref", "change_refs"]


# def pp(s, c: C) -> C:
#    print(s, c)
#    return c

all_methods: Sequence[str] = [
    "get",
    "post",
    "put",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
]


def omit_method_from_path_item(p: PathItem, a: str) -> PathItem:
    return (
        replace(p, get=None)
        if a == "get"
        else replace(p, post=None)
        if a == "post"
        else replace(p, put=None)
        if a == "put"
        else replace(p, delete=None)
        if a == "delete"
        else replace(p, patch=None)
        if a == "patch"
        else replace(p, trace=None)
        if a == "trace"
        else replace(p, options=None)
        if a == "options"
        else replace(p, head=None)
    )


_prism_o = (lambda a: a, lambda b: b)


def odl(
    getter: Callable[[C], D], setter: Callable[[C, D], C]
) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    # TODO: is there a better `None` prism?
    # Seems verbose
    return lens.Lens(getter, setter).Prism(lambda a: a, lambda a: a, ignore_none=True)


def _schema_o_setter(a: C, b: Optional[Schema]):
    return replace(a, schema=b)


schema_o = odl(lambda a: a.schema, _schema_o_setter)


def _request_body_o_setter(a: C, b: Optional[RequestBody]):
    return replace(a, requestBody=b)


request_body_o = odl(lambda a: a.requestBody, _request_body_o_setter)


def _content_o_setter(a: C, b: Optional[Mapping[str, MediaType]]):
    return replace(a, content=b)


content_o = odl(lambda a: a.content, _content_o_setter)


def _paths_o_setter(a: C, b: Optional[Mapping[str, PathItem]]):
    return replace(a, paths=b)


paths_o = odl(lambda a: a.paths, _paths_o_setter)


def _headers_o_setter(a: C, b: Optional[Mapping[str, Union[Header, Reference]]]):
    return replace(a, headers=b)


headers_o = odl(lambda a: a.headers, _headers_o_setter)


def _responses_o(a: C, b: Responses):
    return replace(a, responses=b)


responses_o = odl(lambda a: a.responses, _responses_o)


def _parameters_o_setter(a: C, b: Optional[Sequence[Parameter]]):
    return replace(a, parameters=b)


parameters_o = odl(lambda a: a.parameters, _parameters_o_setter)


def _get_o_setter(a: C, b: Optional[Operation]):
    return replace(a, get=b)


def _post_o_setter(a: C, b: Optional[Operation]):
    return replace(a, post=b)


def _put_o_setter(a: C, b: Optional[Operation]):
    return replace(a, put=b)


def _delete_o_setter(a: C, b: Optional[Operation]):
    return replace(a, delete=b)


def _options_o_setter(a: C, b: Optional[Operation]):
    return replace(a, options=b)


def _head_o_setter(a: C, b: Optional[Operation]):
    return replace(a, head=b)


def _patch_o_setter(a: C, b: Optional[Operation]):
    return replace(a, patch=b)


def _trace_o_setter(a: C, b: Optional[Operation]):
    return replace(a, trace=b)


def operation_o(m: str):
    return {
        "get": odl(lambda a: a.get, _get_o_setter),
        "post": odl(lambda a: a.post, _post_o_setter),
        "put": odl(lambda a: a.put, _put_o_setter),
        "delete": odl(lambda a: a.delete, _delete_o_setter),
        "options": odl(lambda a: a.options, _options_o_setter),
        "head": odl(lambda a: a.head, _head_o_setter),
        "patch": odl(lambda a: a.patch, _patch_o_setter),
        "trace": odl(lambda a: a.trace, _trace_o_setter),
    }[m]


def oll(focus: int) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    """Focus on a specific index in an array if it exists
    """
    return lens.Items().Filter(lambda a: a[0] == focus)[1]


def match_urls(protocol: str, host: str, o: OpenAPIObject) -> Sequence[str]:
    """Finds mock URLs that match a given protocol and host.

    Arguments:
        protocol {str} -- like http or https
        host {str} -- like api.foo.com
        o {OpenAPIObject} -- schema from which the mock URLs are taken

    Returns:
        A list of URLs that match the OpenAPI spec.
    """
    servers = o.servers
    if servers is None:
        return []
    #### if the openapi mock url has no scheme,
    #### we ignore the incoming scheme and treat the url as the host
    #### not sure if this is the right decision,
    #### but it deals with a realistic outcome in certain schemas
    return [
        server.url
        for server in servers
        if (
            (urlparse(server.url).scheme in ["", None])
            or (urlparse(server.url).scheme == protocol)
        )
        and (
            (server.url.split("/")[0] == host) or (urlparse(server.url).netloc == host)
        )
    ]


def get_component_from_ref(
    o: OpenAPIObject,
    d: str,
    accessor: Callable[[Components], Optional[Mapping[str, Union[Reference, C]]]],
    getter: Callable[[OpenAPIObject, Union[C, Reference]], Optional[C]],
) -> Optional[C]:
    return (
        lens.F(lambda a: a.components)
        .F(lambda a: a if a is None else accessor(a))
        .F(lambda a: a if a is None else getter(o, a[d]))
        .get()(o)
    )


def ref_name(r: Reference) -> str:
    return r._ref.split("/")[3]


def internal_get_component(
    f: Callable[[OpenAPIObject, str], Optional[C]]
) -> Callable[[OpenAPIObject, Union[C, Reference]], Optional[C]]:
    def _internal_get_component(
        o: OpenAPIObject, i: Union[C, Reference]
    ) -> Optional[C]:
        return (
            (f(o, ref_name(i)) if isinstance(i, Reference) else i)
            if i is not None
            else None
        )

    return _internal_get_component


# TODO: it seems that there is a bug in pyright that makes get_component_from_ref return a Union[Reference, Schema] instead of Schema
# need to do a cast of the output value to Schema (instead of the union) to fix that
# same for functions below
def get_schema_from_ref(o: OpenAPIObject, d: str) -> Optional[Schema]:
    out = get_component_from_ref(
        o, d, lambda a: a.schemas, internal_get_schema_from_ref
    )
    return None if out is None else cast(Schema, out)


internal_get_schema_from_ref = internal_get_component(get_schema_from_ref)


def get_parameter_from_ref(o: OpenAPIObject, d: str) -> Optional[Parameter]:
    out = get_component_from_ref(
        o, d, lambda a: a.parameters, internal_get_parameter_from_ref
    )
    return None if out is None else cast(Parameter, out)


internal_get_parameter_from_ref = internal_get_component(get_parameter_from_ref)


def get_request_body_from_ref(o: OpenAPIObject, d: str) -> Optional[RequestBody]:
    out = get_component_from_ref(
        o, d, lambda a: a.requestBodies, internal_get_request_body_from_ref
    )
    return None if out is None else cast(RequestBody, out)


internal_get_request_body_from_ref = internal_get_component(get_request_body_from_ref)


def get_response_from_ref(o: OpenAPIObject, d: str) -> Optional[Response]:
    out = get_component_from_ref(
        o, d, lambda a: a.responses, internal_get_response_from_ref
    )
    return None if out is None else cast(Response, out)


internal_get_response_from_ref = internal_get_component(get_response_from_ref)


def schema_prism(oai: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Prism(
        lambda s: get_schema_from_ref(oai, ref_name(s))
        if isinstance(s, Reference)
        else s,
        lambda a: a,
        ignore_none=True,
    )


def change_ref(j: Reference) -> Reference:
    return Reference(_x=None, _ref="#/definitions/%s" % ref_name(j),)


def change_refs(j: Schema) -> Schema:
    return replace(
        j,
        additionalProperties=None
        if j.additionalProperties is None
        else j.additionalProperties
        if isinstance(j.additionalProperties, bool)
        else change_ref(j.additionalProperties)
        if isinstance(j.additionalProperties, Reference)
        else change_refs(j.additionalProperties),
        items=None
        if j.items is None
        else change_ref(j.items)
        if isinstance(j.items, Reference)
        else change_refs(j.items)
        if isinstance(j.items, Schema)
        else [
            change_ref(item) if isinstance(item, Reference) else change_refs(item)
            for item in j.items
        ],
        _not=None
        if j._not is None
        else change_ref(j._not)
        if isinstance(j._not, Reference)
        else change_refs(j._not),
        anyOf=None
        if j.anyOf is None
        else [
            change_ref(item) if isinstance(item, Reference) else change_refs(item)
            for item in j.anyOf
        ],
        allOf=None
        if j.allOf is None
        else [
            change_ref(item) if isinstance(item, Reference) else change_refs(item)
            for item in j.allOf
        ],
        oneOf=None
        if j.oneOf is None
        else [
            change_ref(item) if isinstance(item, Reference) else change_refs(item)
            for item in j.oneOf
        ],
        properties=None
        if j.properties is None
        else {
            k: change_ref(v) if isinstance(v, Reference) else change_refs(v)
            for k, v in j.properties.items()
        },
    )


def make_definitions_from_schema(o: OpenAPIObject) -> Mapping[str, Any]:
    return (
        {
            a: convert_from_openapi(
                change_ref(b) if isinstance(b, Reference) else change_refs(b)
            )
            for a, b in o.components.schemas.items()
        }
        if o.components and o.components.schemas
        else {}
    )


def find_relevant_path(m: str, a: Sequence[str], p: PathItem) -> PathItem:
    return (
        p
        if len(a) == 0
        else find_relevant_path(
            m, a[1:], p if a[0] == m else omit_method_from_path_item(p, a[0]),
        )
    )


def get_path_item_with_method(m: str, p: PathItem) -> PathItem:
    return find_relevant_path(m, all_methods, p)


def get_required_request_query_or_header_parameters_internal(
    header: bool, l: lenses.ui.BaseUiLens[S, T, X, Y], oai: OpenAPIObject, p: PathItem
) -> Sequence[Parameter]:
    # TODO: really dirty cast
    # is this even the case
    # copied from unmock, but need to investigate
    return [
        replace(
            parameter,
            name=parameter.name if not header else parameter.name.lower(),
            schema=change_ref(parameter.schema)
            if isinstance(parameter.schema, Reference)
            else change_refs(parameter.schema)
            if parameter.schema is not None
            else Schema(_type="string"),
        )
        for parameter in l.Prism(
            lambda s: get_parameter_from_ref(oai, ref_name(s))
            if isinstance(s, Reference)
            else s,
            lambda a: a,
            ignore_none=True,
        )
        .Prism(
            lambda s: s
            if (s._in == ("header" if header else "query")) and s.required
            else None,
            lambda a: a,
            ignore_none=True,
        )
        .collect()(p)
        if parameter.schema is not None
        and (
            len(lens.Each().add_lens(schema_prism(oai)).collect()([parameter.schema]))
            > 0
        )
    ]


def get_required_request_query_or_header_parameters(
    header: bool, req: Request, oai: OpenAPIObject, p: PathItem
) -> Sequence[Parameter]:
    return [
        *get_required_request_query_or_header_parameters_internal(
            header, parameters_o.Each(), oai, p
        ),
        *get_required_request_query_or_header_parameters_internal(
            header, operation_o(req.method.value).add_lens(parameters_o).Each(), oai, p
        ),
    ]


def get_required_request_body_schemas(
    req: Request, oai: OpenAPIObject, p: PathItem,
) -> Sequence[Union[Reference, Schema]]:
    return (
        operation_o(req.method.value)
        .add_lens(request_body_o)
        .Prism(
            lambda s: get_request_body_from_ref(oai, ref_name(s))
            if isinstance(s, Reference)
            else s,
            lambda a: a,
            ignore_none=True,
        )
        # automatically ignore not required for now
        .Prism(
            lambda s: None if s.required is False else s, lambda a: a, ignore_none=True
        )
        .add_lens(content_o)
        .Values()
        .add_lens(schema_o)
        .add_lens(schema_prism(oai))
        .collect()(p)
    )


def valid_schema(to_validate: Any, schema: Any) -> bool:
    try:
        jsonschema.validate(to_validate, schema)
        return True
    except Exception:
        # print(e)
        return False


def keep_method_if_required_request_body_is_present(
    req: Request, oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_request_body_is_present(p: PathItem) -> PathItem:
        out = (
            p
            if (operation_from_string(p, req.method.value) is None)
            or (
                len(
                    [
                        s
                        for s in get_required_request_body_schemas(req, oai, p)
                        if not valid_schema(
                            req.bodyAsJson
                            if req.bodyAsJson is not None
                            else json.loads(req.body)
                            if req.body is not None
                            else "",
                            {
                                # TODO: this line is different than the TS implementation
                                # because I think there is a logic bug there
                                # it should look like this line as we are not sure
                                # if the schema will be a reference or not
                                # perhaps I'm wrong in the assumption... only testing will tell...
                                # otherwise, change the TS implementation in unmock-js and delete this comment.
                                **convert_from_openapi(
                                    change_ref(s)
                                    if isinstance(s, Reference)
                                    else change_refs(s)
                                ),
                                "definitions": make_definitions_from_schema(oai),
                            },
                        )
                    ]
                )
                == 0
            )
            else omit_method_from_path_item(p, req.method.value)
        )
        return out

    return _keep_method_if_required_request_body_is_present


def keep_method_if_required_header_parameters_are_present(
    req: Request, oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_header_parameters_are_present(p: PathItem) -> PathItem:
        out = keep_method_if_required_query_or_header_parameters_are_present(
            True, req, oai, p
        )
        return out

    return _keep_method_if_required_header_parameters_are_present


def keep_method_if_required_query_parameters_are_present(
    req: Request, oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_query_parameters_are_present(p: PathItem) -> PathItem:
        out = keep_method_if_required_query_or_header_parameters_are_present(
            False, req, oai, p
        )
        return out

    return _keep_method_if_required_query_parameters_are_present


def _json_schema_from_required_parameters(
    parameters: Sequence[Parameter], oai: OpenAPIObject
) -> Any:
    return {
        "type": "object",
        "properties": {
            param.name: convert_from_openapi(param.schema) for param in parameters
        },
        "required": [param.name for param in parameters if param.required],
        "additionalProperties": True,
        "definitions": make_definitions_from_schema(oai),
    }


def keep_method_if_required_query_or_header_parameters_are_present(
    header: bool, req: Request, oai: OpenAPIObject, p: PathItem
) -> PathItem:
    return (
        p
        if (operation_from_string(p, req.method.value) is None)
        or valid_schema(
            {k.lower(): v for k, v in req.headers.items()} if header else req.query,
            _json_schema_from_required_parameters(
                get_required_request_query_or_header_parameters(header, req, oai, p),
                oai,
            ),
        )
        else omit_method_from_path_item(p, req.method.value)
    )


def maybe_add_string_schema(
    s: Sequence[Union[Reference, Schema]]
) -> Sequence[Union[Reference, Schema]]:
    return [Schema(_type="string")] if len(s) == 0 else s


def discern_name(o: Optional[Parameter], n: str) -> Optional[Parameter]:
    return o if (o is None) or ((o.name == n) and (o._in == "path")) else None


def internal_get_parameter_schemas(
    t: lenses.ui.BaseUiLens[S, T, X, Y],
    vname: str,
    path_item: PathItem,
    oas: OpenAPIObject,
) -> Sequence[Schema]:
    return (
        t.Prism(
            lambda i: discern_name(get_parameter_from_ref(oas, ref_name(i)), vname)
            if isinstance(i, Reference)
            else discern_name(i, vname),
            lambda a: a,
            ignore_none=True,
        )
        .add_lens(schema_o)
        .collect()(path_item)
    )


def path_item_path_parameter_schemas(
    vname: str, path_item: PathItem, oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(parameters_o.Each(), vname, path_item, oas,)


def operation_path_parameter_schemas(
    vname: str, path_item: PathItem, operation: str, oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(
        operation_o(operation.lower()).add_lens(parameters_o).Each(),
        vname,
        path_item,
        oas,
    )


def get_matching_parameters(
    vname: str, path_item: PathItem, operation: str, oas: OpenAPIObject,
) -> Sequence[Union[Schema, Reference]]:
    return maybe_add_string_schema(
        [
            *path_item_path_parameter_schemas(vname, path_item, oas),
            *operation_path_parameter_schemas(vname, path_item, operation, oas),
        ]
    )


def maybeJson(maybe: str) -> Any:
    try:
        return json.loads(maybe)
    except json.JSONDecodeError:
        return maybe


def path_parameter_match(
    part: str, vname: str, path_item: PathItem, operation: str, oas: OpenAPIObject,
) -> bool:
    """Matches part of a path against a path parameter with name vname

    Arguments:
        part {str} -- part of a path, ie an id
        vname {str} -- name of a parameter
        path_item {PathItem} -- a path item maybe containing the parameter
        operation {MethodNames} -- the name of the operation to check in case the parameter is in the operation
        oas {OpenAPIObject} -- the schema to traverse to find definitions

    Returns:
        bool -- Did we get a match
    """
    return reduce(
        lambda q, r: q
        and reduce(
            lambda a, b: a
            or valid_schema(
                b,
                {
                    **convert_from_openapi(r),
                    "definitions": make_definitions_from_schema(oas),
                },
            ),
            list(set([part, maybeJson(part)])),
            False,
        ),
        get_matching_parameters(vname, path_item, operation, oas),
        True,
    )


path_param_regex = re.compile(r"^\{[^}]+\}")


# TODO: this is a boolean nightmare
# what we want is
# lengths_are_equal AND (empty OR ((first_elt_equal OR (matches_regex AND matches_schema)) AND recursion))
def matches_internal(
    path: Sequence[str],
    path_item_key: Sequence[str],
    path_item: PathItem,
    operation: str,
    o: OpenAPIObject,
) -> bool:
    return (len(path) == len(path_item_key)) and (
        (len(path) == 0)
        or (
            (
                (path[0] == path_item_key[0])
                or (
                    (path_param_regex.match(path_item_key[0]) is not None)
                    and path_parameter_match(
                        path[0], path_item_key[0][1:-1], path_item, operation, o,
                    )
                )
            )
            and matches_internal(path[1:], path_item_key[1:], path_item, operation, o,)
        )
    )


def matches(
    path: str, path_item_key: str, path_item: PathItem, method: str, oas: OpenAPIObject,
) -> bool:
    return matches_internal(
        [x for x in path.split("/") if x != ""],
        [x for x in path_item_key.split("/") if x != ""],
        path_item,
        method,
        oas,
    )


def get_first_method_internal_2(
    p: PathItem, n: str, m: Sequence[str], o: Optional[Operation],
) -> Optional[Tuple[str, Operation]]:
    return (n, o) if o is not None else get_first_method_internal(m, p)


def get_first_method_internal(
    m: Sequence[str], p: PathItem,
) -> Optional[Tuple[str, Operation]]:
    return (
        None
        if len(m) == 0
        else get_first_method_internal_2(p, m[0], m[1:], operation_from_string(p, m[0]))
    )


def get_first_method(p: PathItem) -> Optional[Tuple[str, Operation]]:
    return get_first_method_internal(all_methods, p)


def get_header_from_ref(o: OpenAPIObject, d: str) -> Optional[Header]:
    out = get_component_from_ref(
        o, d, lambda a: a.headers, internal_get_header_from_ref,
    )
    return None if out is None else cast(Header, out)


internal_get_header_from_ref = internal_get_component(get_header_from_ref)


def use_if_header_last_mile(
    p: Parameter, r: Optional[Schema],
) -> Optional[Tuple[str, Schema]]:
    return (
        None if r is None else (p.name, r if r is not None else Schema(_type="string"))
    )


def use_if_header(o: OpenAPIObject, p: Parameter,) -> Optional[Tuple[str, Schema]]:
    return (
        None
        if p._in != "header"
        else use_if_header_last_mile(
            p,
            Schema(_type="string")
            if p.schema is None
            else get_schema_from_ref(o, ref_name(p.schema))
            if isinstance(p.schema, Reference)
            else p.schema,
        )
    )


def parameter_schema(o: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Iso(lambda a: (use_if_header(o, a), a), lambda b: b[1])[0].Prism(
        *_prism_o, ignore_none=True
    )


def cut_path(paths: Sequence[str], path: str) -> str:
    return (
        path
        if len(paths) == 0
        else path[len(paths[0]) :]
        if path[: len(paths[0])] == paths[0]
        else cut_path(paths[1:], path)
    )


def remove_trailing_slash(s: str) -> str:
    return s if len(s) == 0 else s[:-1] if s[-1] == "/" else s


def truncate_path(path: str, o: OpenAPIObject, i: Request,) -> str:
    return cut_path(
        [
            remove_trailing_slash(urlparse(u).path)
            for u in match_urls(i.protocol.value, i.host, o)
        ],
        path,
    )


def match_request_to_openapi(
    req: Request, specs: Sequence[OpenAPISpecification]
) -> Sequence[OpenAPISpecification]:
    def _path_item_modifier(oai: OpenAPIObject) -> Callable[[PathItem], PathItem]:
        def __path_item_modifier(path_item: PathItem) -> PathItem:
            return reduce(
                lambda a, b: b(a),
                [
                    keep_method_if_required_header_parameters_are_present(req, oai),
                    keep_method_if_required_query_parameters_are_present(req, oai),
                    keep_method_if_required_request_body_is_present(req, oai),
                ],
                get_path_item_with_method(req.method.value, path_item),
            )

        return __path_item_modifier

    def _oai_modifier(oai: OpenAPIObject) -> OpenAPIObject:
        return paths_o.Values().modify(_path_item_modifier(oai))(
            replace(
                oai,
                paths={
                    n: o
                    for n, o in oai.paths.items()
                    if matches(
                        truncate_path(req.pathname, oai, req),
                        n,
                        o,
                        req.method.value,
                        oai,
                    )
                },
            )
        )

    specs_with_matching_urls = {
        spec.source: spec.api
        for spec in specs
        if len(match_urls(req.protocol.value, req.host, spec.api)) > 0
    }

    filtered = {k: _oai_modifier(v) for k, v in specs_with_matching_urls.items()}

    return [OpenAPISpecification(api, source) for (source, api) in filtered.items()]



import json
from typing import Sequence

from http_types import RequestBuilder
from openapi_typed_2 import convert_to_openapi

from meeshkan.serve.mock.matcher import match_request_to_openapi
from meeshkan.serve.mock.specs import OpenAPISpecification

store: Sequence[OpenAPISpecification] = [
    OpenAPISpecification(
        convert_to_openapi(
            {
                "openapi": "",
                "servers": [
                    {"url": "api.foo.com"}
                ],  # we omit the protocol and it should still match
                "info": {"title": "", "version": ""},
                "paths": {
                    "/user": {
                        "get": {"responses": {"200": {"description": "userget"}}},
                        "post": {"responses": {"200": {"description": "userpost"}}},
                        "description": "",
                    },
                    "/user/{id}": {
                        "parameters": [
                            {
                                "name": "id",
                                "in": "path",
                                "required": True,
                                "schema": {"type": "integer"},
                            },
                        ],
                        "get": {"responses": {"200": {"description": "useridget"}}},
                        "post": {"responses": {"201": {"description": "useridpost"}}},
                    },
                    "/user/{id}/name": {
                        "get": {"responses": {"200": {"description": "useridnameget"}}},
                        "post": {
                            "responses": {"201": {"description": "useridnamepost"}}
                        },
                        "description": "",
                    },
                },
            }
        ),
        "foo",
    ),
    OpenAPISpecification(
        convert_to_openapi(
            {
                "openapi": "",
                "servers": [{"url": "https://api.bar.com/v1"}],
                "info": {"title": "", "version": ""},
                "paths": {
                    "/guest": {
                        "get": {"responses": {"200": {"description": "guestget"}}},
                        "post": {"responses": {"200": {"description": "guestpost"}}},
                        "description": "",
                    },
                    "/guest/{id}": {
                        "get": {"responses": {"200": {"description": "guestidget"}}},
                        "post": {"responses": {"201": {"description": "guestidpost"}}},
                        "description": "",
                    },
                    "/guest/{id}/name": {
                        "get": {
                            "responses": {"200": {"description": "guestidnameget"}}
                        },
                        "post": {
                            "responses": {"201": {"description": "guestidnamepost"}}
                        },
                        "description": "",
                    },
                },
            }
        ),
        "bar",
    ),
    OpenAPISpecification(
        convert_to_openapi(
            {
                "openapi": "",
                "servers": [{"url": "https://api.baz.com"}],
                "info": {"title": "", "version": ""},
                "paths": {
                    "/guest": {
                        "parameters": [
                            {
                                "in": "query",
                                "required": True,
                                "name": "hello",
                                "schema": {"type": "string", "pattern": "^[0-9]+$"},
                            },
                        ],
                        "get": {"responses": {"200": {"description": "guestget"}}},
                        "post": {
                            "parameters": [
                                {
                                    "in": "query",
                                    "required": True,
                                    "name": "thinking",
                                    "schema": {"type": "string"},
                                },
                            ],
                            "responses": {"200": {"description": "guestpost"}},
                        },
                        "description": "",
                    },
                    "/guest/{id}": {
                        "parameters": [
                            {
                                "in": "query",
                                "required": True,
                                "name": "a",
                                "schema": {"type": "string"},
                            },
                            {
                                "in": "query",
                                "required": True,
                                "name": "b",
                                "schema": {"type": "string"},
                            },
                            {"in": "query", "name": "c", "schema": {"type": "string"}},
                        ],
                        "get": {"responses": {"200": {"description": "guestidget"}}},
                        "post": {
                            "parameters": [
                                {
                                    "in": "header",
                                    "required": True,
                                    "name": "zz",
                                    "schema": {"type": "string"},
                                },
                                {
                                    "in": "query",
                                    "required": True,
                                    "name": "zzz",
                                    "schema": {"type": "string"},
                                },
                                {
                                    "in": "query",
                                    "name": "c",
                                    "schema": {"type": "string"},
                                },
                            ],
                            "responses": {"201": {"description": "guestidpost"}},
                        },
                        "description": "",
                    },
                    "/guest/{id}/name": {
                        "get": {
                            "responses": {"200": {"description": "guestidnameget"}}
                        },
                        "post": {
                            "requestBody": {
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "required": ["age"],
                                            "properties": {"age": {"type": "integer"}},
                                        },
                                    },
                                },
                            },
                            "responses": {"201": {"description": "guestidnamepost"}},
                        },
                        "description": "",
                    },
                },
            }
        ),
        "baz",
    ),
]


def test_matcher_1():
    assert match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.foo.com",
                "path": "/user",
                "pathname": "/user",
                "protocol": "https",
                "method": "get",
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "api.foo.com"}],
                    "info": {"title": "", "version": ""},
                    "paths": {
                        "/user": {
                            "get": {"responses": {"200": {"description": "userget"}}},
                            "description": "",
                        },
                    },
                }
            ),
            "foo",
        )
    ]


def test_matcher_2():
    match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.bar.com",
                "path": "/v1/guest/{id}",
                "pathname": "/v1/guest/{id}",
                "protocol": "https",
                "method": "post",
                "query": {},
            }
        ),
        store,
    ) == {
        "bar": convert_to_openapi(
            {
                "openapi": "",
                "servers": [{"url": "https://api.bar.com/v1"}],
                "info": {"title": "", "version": ""},
                "paths": {
                    "/guest/{id}": {
                        "post": {"responses": {"201": {"description": "guestidpost"}}},
                        "description": "",
                    },
                },
            }
        )
    }


def test_matcher_3():
    assert match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.foo.com",
                "path": "/users",  # incorrect, should be user
                "pathname": "/users",  # incorrect, should be user
                "protocol": "https",
                "method": "get",
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "api.foo.com"}],
                    "info": {"title": "", "version": ""},
                    "paths": {},
                }
            ),
            "foo",
        )
    ]


def test_matcher_4():
    assert match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.foo.com",
                "path": "/user/55",  # correctly parses number
                "pathname": "/user/55",  # correcly parses number
                "protocol": "https",
                "method": "get",
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "api.foo.com"}],
                    "info": {"title": "", "version": ""},
                    "paths": {
                        "/user/{id}": {
                            "parameters": [
                                {
                                    "name": "id",
                                    "in": "path",
                                    "required": True,
                                    "schema": {"type": "integer"},
                                },
                            ],
                            "get": {"responses": {"200": {"description": "useridget"}}},
                        },
                    },
                }
            ),
            "foo",
        ),
    ]


def test_matcher_5():
    assert match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.foo.com",
                "path": "/user/fdsfsfwef",  # correctly rejects non-number
                "pathname": "/user/fdsfsfwef",  # correcly rejects non-number
                "protocol": "https",
                "method": "get",
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "api.foo.com"}],
                    "info": {"title": "", "version": ""},
                    "paths": {},
                }
            ),
            "foo",
        )
    ]


def test_matcher_6():
    assert match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {},
                "host": "api.foo.com",
                "path": "/user",
                "pathname": "/user",
                "protocol": "https",
                "method": "delete",  # does not exist
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "api.foo.com"}],
                    "info": {"title": "", "version": ""},
                    "paths": {"/user": {"description": ""}},
                }
            ),
            "foo",
        )
    ]


def test_matcher_7():
    assert (
        match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.foo.commmm",  # does not exist
                    "path": "/user",
                    "pathname": "/user",
                    "protocol": "https",
                    "method": "get",
                    "query": {},
                }
            ),
            store,
        )
        == []
    )


def test_matcher_8():
    (baz_spec,) = (spec for spec in store if spec.source == "baz")
    (matched_spec,) = (
        spec
        for spec in match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.baz.com",
                    "path": "/guest",
                    "pathname": "/guest",
                    "protocol": "https",
                    "method": "get",
                    "query": {"hello": "0"},  # query param must be integer
                }
            ),
            store,
        )
        if spec.source == "baz"
    )
    assert matched_spec.api.paths["/guest"].get == baz_spec.api.paths["/guest"].get


def test_matcher_9():
    (matched_spec,) = (
        spec
        for spec in match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.baz.com",
                    "path": "/guest",
                    "pathname": "/guest",
                    "protocol": "https",
                    "method": "get",
                    "query": {"hello": "b"},  # query param intentionally off as string
                }
            ),
            store,
        )
        if spec.source == "baz"
    )
    assert matched_spec.api.paths["/guest"].get is None


def test_matcher_10():
    assert (
        match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.baz.com",
                    "path": "/guest",
                    "pathname": "/guest",
                    "protocol": "https",
                    "method": "get",
                    "query": {},  # as query is required, this should fail
                }
            ),
            store,
        )[0]
        .api.paths["/guest"]
        .get
        is None
    )


def test_matcher_11():
    (baz_spec,) = (spec for spec in store if spec.source == "baz")
    (matched_spec,) = (
        spec
        for spec in match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.baz.com",
                    "path": "/guest/3/name",
                    "pathname": "/guest/3/name",
                    "protocol": "https",
                    "method": "post",
                    "query": {},
                    "bodyAsJson": {"age": 42},
                    "body": json.dumps({"age": 42}),
                }
            ),
            store,
        )
        if spec.source == "baz"
    )
    assert (
        matched_spec.api.paths["/guest/{id}/name"].post
        == baz_spec.api.paths["/guest/{id}/name"].post
    )


def test_matcher_12():
    assert (
        match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},
                    "host": "api.baz.com",
                    "path": "/guest/3/name",
                    "pathname": "/guest/3/name",
                    "protocol": "https",
                    "method": "post",
                    "query": {},
                    "body": json.dumps({"age": "42"}),
                    "bodyAsJson": {"age": "42"},  # wrong type, as 42 is string
                }
            ),
            store,
        )[0]
        .api.paths["/guest/{id}/name"]
        .post
        is None
    )


def test_matcher_13():
    (baz_spec,) = (spec for spec in store if spec.source == "baz")
    (matched_spec,) = (
        spec
        for spec in match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {"zz": "top"},
                    "host": "api.baz.com",
                    "path": "/guest/4",
                    "pathname": "/guest/4",
                    "protocol": "https",
                    "method": "post",
                    "query": {"zzz": "aaa", "a": "foo", "b": "baz"},
                }
            ),
            store,
        )
        if spec.source == "baz"
    )
    assert (
        matched_spec.api.paths["/guest/{id}"].post
        == baz_spec.api.paths["/guest/{id}"].post
    )


def test_matcher_14():
    assert (
        match_request_to_openapi(
            RequestBuilder.from_dict(
                {
                    "headers": {},  # no header will lead to undefined
                    "host": "api.baz.com",
                    "path": "/guest/4",
                    "pathname": "/guest/4",
                    "protocol": "https",
                    "method": "post",
                    "query": {"zzz": "aaa", "a": "foo", "b": "baz"},
                }
            ),
            store,
        )[0]
        .api.paths["/guest/{id}"]
        .post
        is None
    )


def test_matcher_15():
    match_request_to_openapi(
        RequestBuilder.from_dict(
            {
                "headers": {"X-Meeshkan-Scheme": "https", "Host": "api.bar.com"},
                "host": "api.bar.com",
                "path": "/v1/guest/{id}",
                "pathname": "/v1/guest/{id}",
                "protocol": "http",
                "method": "post",
                "query": {},
            }
        ),
        store,
    ) == [
        OpenAPISpecification(
            convert_to_openapi(
                {
                    "openapi": "",
                    "servers": [{"url": "https://api.bar.com/v1"}],
                    "info": {"title": "", "version": ""},
                    "paths": {
                        "/guest/{id}": {
                            "post": {
                                "responses": {"201": {"description": "guestidpost"}}
                            },
                            "description": "",
                        },
                    },
                }
            ),
            "baz",
        )
    ]



import logging

from tornado.httpserver import HTTPServer
from tornado.web import Application

from meeshkan.serve.admin.views import (
    RestMiddlewaresView,
    RestMiddlewareView,
    ScopeView,
    StorageView,
)
from meeshkan.serve.mock.rest import rest_middleware_manager
from meeshkan.serve.mock.storage import storage_manager

from ..mock.scope import Scope

logger = logging.getLogger(__name__)


def make_admin_app(
    scope,
    storage_manager=storage_manager,
    rest_middleware_manager=rest_middleware_manager,
):
    storage_view_deps = dict(storage_manager=storage_manager)
    rest_middleware_deps = dict(rest_middleware_manager=rest_middleware_manager)
    scope_view_deps = dict(scope=scope)
    return Application(
        [
            (r"/admin/scope", ScopeView, scope_view_deps),
            (r"/admin/storage", StorageView, storage_view_deps),
            (
                r"/admin/middleware/rest/pregen",
                RestMiddlewaresView,
                rest_middleware_deps,
            ),
            (
                r"/admin/middleware/rest/pregen/(.+)",
                RestMiddlewareView,
                rest_middleware_deps,
            ),
        ]
    )


def start_admin(port, scope: Scope):
    app = make_admin_app(scope)
    http_server = HTTPServer(app)
    http_server.listen(port)
    logger.info("- Admin   http://localhost:%s/admin", port)



import json
from unittest.mock import Mock

import pytest
from http_types import HttpMethod, Protocol
from http_types.utils import ResponseBuilder
from tornado.httpclient import HTTPRequest

from meeshkan.serve.mock.callbacks import callback_manager
from meeshkan.serve.mock.log import Log
from meeshkan.serve.mock.response_matcher import ResponseMatcher
from meeshkan.serve.mock.scope import Scope
from meeshkan.serve.mock.server import make_mocking_app_
from meeshkan.serve.utils.routing import HeaderRouting

response = ResponseBuilder.from_dict(
    dict(
        statusCode=200,
        body='{"message": "hello"}',
        bodyAsJson=json.loads('{"message": "hello"}'),
        headers={},
    )
)
get_response_mock = Mock(return_value=response)


@pytest.fixture
def app():
    callback_manager.load("tests/serve/mock/callbacks")
    response_matcher = ResponseMatcher("tests/serve/mock/schemas/petstore")

    response_matcher.get_response = get_response_mock

    app = make_mocking_app_(
        callback_manager=callback_manager,
        response_matcher=response_matcher,
        router=HeaderRouting(),
        log=Log(Scope()),
    )
    return app


@pytest.mark.gen_test
def test_mocking_server_pets(http_client, base_url, app):

    req = HTTPRequest(base_url + "/pets", headers={"Host": "petstore.swagger.io"})
    http_response = yield http_client.fetch(req)
    assert 200 == http_response.code
    rb = json.loads(http_response.body)
    assert {"message": "hello"} == rb

    assert len(get_response_mock.call_args_list) == 1
    request = get_response_mock.call_args_list[0][0][0]
    assert HttpMethod.GET == request.method
    assert Protocol.HTTP == request.protocol
    assert "/pets" == request.pathname
    assert "/pets" == request.path
    assert {} == request.query
    assert "petstore.swagger.io" == request.host
    assert "petstore.swagger.io" == request.headers["Host"]



import logging

import tornado.ioloop
from http_types import Request, Response

from meeshkan.serve.utils.data_callback import RequestLoggingCallback
from meeshkan.serve.utils.routing import HeaderRouting

from .channel import Channel
from .proxy_callback import ProxyCallback

logger = logging.getLogger(__name__)


class ProxyBase(ProxyCallback):
    def __init__(self, data_callback, router, ssl_options=None):
        super().__init__(ssl_options=ssl_options)
        self._data_callback = data_callback
        self._channels = dict()
        self._router = router

    def handle_stream(self, stream, address):
        logger.debug(
            "Creating channel for new connection from client {}".format(address)
        )
        self._channels[address] = self._create_channel(stream, address)

    def on_remove_channel(self, client_address):
        channel = self._channels[client_address]
        del self._channels[client_address]
        del channel

    def _create_channel(self, stream, client_address):
        raise NotImplementedError()

    def on_request_complete(self, request: Request, response: Response):
        pass

    @property
    def router(self):
        return self._router


class RecordProxy(ProxyBase):
    def __init__(self, data_callback, router, ssl_options=None):
        super().__init__(data_callback, router, ssl_options=ssl_options)

    def _create_channel(self, stream, client_address):
        return Channel(self, stream, client_address, self.router)

    def on_request_complete(self, request: Request, response: Response):
        self._data_callback.log(request, response)


class RecordProxyRunner:
    def __init__(
        self, port, log_dir, specs_dir, routing=HeaderRouting(), mode=None,
    ):
        self._port = port
        self._log_dir = log_dir
        self._specs_dir = specs_dir
        self._routing = routing
        self._mode = mode

    def run(self):

        logger.info("Starting Meeshkan record on http://localhost:%s", self._port)
        logger.info(
            "Spec generation mode is %s",
            self._mode.name.lower() if self._mode else "disabled",
        )
        with RequestLoggingCallback(
            log_dir=self._log_dir, specs_dir=self._specs_dir, update_mode=self._mode
        ) as callback:
            server = RecordProxy(callback, self._routing)
            server.listen(self._port)
            logger.info("Meeshkan is running")
            tornado.ioloop.IOLoop.instance().start()
